from collections.abc import Iterable
from functools import cache
from math import isqrt

import numpy as np
import numpy.typing as npt

from .. import abc
from ..types import Array1D, Array2D
from .PAMConstellation import PAMConstellation


class QAMConstellation(abc.Constellation[np.complexfloating]):
    r"""
    Quadrature amplitude modulation (QAM) constellation. It is a complex one-dimensional [constellation](/ref/Constellation) obtained by a Cartesian product of two [PAM constellations](/ref/PAMConstellation), namely, the *in-phase constellation*, and the *quadrature constellation*. For more details, see <cite>SA15, Sec. 2.5.1</cite>.

    Parameters:
        orders: A tuple $(M_\mathrm{I}, M_\mathrm{Q})$ with the orders of the in-phase and quadrature constellations, respectively. If specified as a single integer $M$, then it is assumed that $M_\mathrm{I} = M_\mathrm{Q} = \sqrt{M}$; in this case, $M$ must be a perfect square.

        deltas: A tuple $(\Delta_\mathrm{I}, \Delta_\mathrm{Q})$ with the distances of the in-phase and quadrature constellations, respectively. If specified as a single float $\Delta$, then it is assumed that $\Delta_\mathrm{I} = \Delta_\mathrm{Q} = \Delta$. The default value is `2.0`.

        phase_offset: The phase offset $\phi$ of the constellation (in turns, not radians). The default value is `0.0`.

    Examples:
        1. The square $16$-QAM constellation with $(M_\mathrm{I}, M_\mathrm{Q}) = (4, 4)$ and $(\Delta_\mathrm{I}, \Delta_\mathrm{Q}) = (2, 2)$ is depicted below.
            <figure markdown>
            ![16-QAM constellation.](/fig/constellation_qam_16.svg)
            </figure>

                >>> const = komm.QAMConstellation(16)

        1.  The rectangular $8$-QAM constellation with $(M_\mathrm{I}, M_\mathrm{Q}) = (4, 2)$ and $(\Delta_\mathrm{I}, \Delta_\mathrm{Q}) = (2, 4)$ is depicted below.
            <figure markdown>
            ![8-QAM constellation.](/fig/constellation_qam_8.svg)
            </figure>

                >>> const = komm.QAMConstellation(orders=(4, 2), deltas=(2.0, 4.0))
    """

    _orders: tuple[int, int]
    _deltas: tuple[float, float]
    phase_offset: float

    def __init__(
        self,
        orders: tuple[int, int] | int,
        deltas: tuple[float, float] | float = 2.0,
        phase_offset: float = 0.0,
    ) -> None:
        if isinstance(orders, Iterable):
            self._orders = orders
        else:
            if not isqrt(orders) ** 2 == orders:
                raise ValueError(
                    "when a single integer, 'orders' must be a perfect square"
                )
            self._orders = (isqrt(orders), isqrt(orders))
        if isinstance(deltas, Iterable):
            self._deltas = deltas
        else:
            self._deltas = (deltas, deltas)
        self._phase_offset = phase_offset

    def __repr__(self) -> str:
        args = ", ".join([
            f"orders={tuple(map(int, self._orders))}",
            f"deltas={tuple(map(float, self._deltas))}",
            f"phase_offset={self._phase_offset}",
        ])
        return f"{self.__class__.__name__}({args})"

    @property
    @cache
    def matrix(self) -> Array2D[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.matrix
            array([[-3.-3.j],
                   [-3.-1.j],
                   [-3.+1.j],
                   [-3.+3.j],
                   [-1.-3.j],
                   [-1.-1.j],
                   [-1.+1.j],
                   [-1.+3.j],
                   [ 1.-3.j],
                   [ 1.-1.j],
                   [ 1.+1.j],
                   [ 1.+3.j],
                   [ 3.-3.j],
                   [ 3.-1.j],
                   [ 3.+1.j],
                   [ 3.+3.j]])
        """
        Mi, Mq = self._orders
        Δi, Δq = self._deltas
        φ = self._phase_offset
        matrix_i = PAMConstellation(Mi, Δi).matrix
        matrix_q = PAMConstellation(Mq, Δq).matrix
        matrix = (matrix_i + 1j * matrix_q.T) * np.exp(2j * np.pi * φ)
        return matrix.reshape(-1, 1)

    @property
    def order(self) -> int:
        r"""
        For the QAM constellation, it is given by
        $$
            M = M_\mathrm{I} M_\mathrm{Q}.
        $$

        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.order
            16
        """
        Mi, Mq = self._orders
        return Mi * Mq

    @property
    def dimension(self) -> int:
        r"""
        For the QAM constellation, it is given by $N = 1$.

        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.dimension
            1
        """
        return 1

    def mean(self, priors: npt.ArrayLike | None = None) -> Array1D[np.complexfloating]:
        r"""
        For uniform priors, the mean of the QAM constellation is given by
        $$
            \mathbf{m} = 0.
        $$

        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.mean()
            array([0.+0.j])
        """
        if priors is None:
            return np.zeros((1,), dtype=complex)
        return super().mean(priors)

    def mean_energy(self, priors: npt.ArrayLike | None = None) -> np.floating:
        r"""
        For uniform priors, the mean energy of the QAM constellation is given by
        $$
            E = \frac{\Delta_\mathrm{I}^2}{12}(M_\mathrm{I}^2 - 1) + \frac{\Delta_\mathrm{Q}^2}{12}(M_\mathrm{Q}^2 - 1).
        $$
        For the special case of a square QAM constellation, this simplifies to
        $$
            E = \frac{\Delta^2}{6}(M - 1).
        $$

        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.mean_energy()
            np.float64(10.0)
        """
        if priors is None:
            Mi, Mq = self._orders
            Δi, Δq = self._deltas
            return np.float64(Δi**2 * (Mi**2 - 1) / 12 + Δq**2 * (Mq**2 - 1) / 12)
        return super().mean_energy(priors)

    def minimum_distance(self) -> np.floating:
        r"""
        For the QAM constellation, the minimum distance is given by
        $$
            d_{\min} = \min(\Delta_\mathrm{I}, \Delta_\mathrm{Q}).
        $$

        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.minimum_distance()
            np.float64(2.0)
        """
        return np.min(self._deltas)

    def indices_to_symbols(
        self, indices: npt.ArrayLike
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.indices_to_symbols([3, 0])
            array([-3.+3.j, -3.-3.j])
            >>> const.indices_to_symbols([[3, 0], [1, 2]])
            array([[-3.+3.j, -3.-3.j],
                   [-3.-1.j, -3.+1.j]])
        """
        return super().indices_to_symbols(indices)

    def closest_indices(self, received: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.closest_indices([-3.1 + 2.9j, -3 - 3.5j])
            array([3, 0])
        """
        Mi, Mq = self._orders
        Δi, Δq = self._deltas
        φ = self._phase_offset
        Ai, Aq = Δi / 2, Δq / 2
        received_i = np.real(np.multiply(received, np.exp(-2j * np.pi * φ))) / Ai
        received_q = np.imag(np.multiply(received, np.exp(-2j * np.pi * φ))) / Aq
        indices_i = np.clip(np.around((received_i + Mi - 1) / 2), 0, Mi - 1).astype(int)
        indices_q = np.clip(np.around((received_q + Mq - 1) / 2), 0, Mq - 1).astype(int)
        indices = Mq * indices_i + indices_q
        return indices

    def closest_symbols(
        self, received: npt.ArrayLike
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.closest_symbols([-3.1 + 2.9j, -3 - 3.5j])
            array([-3.+3.j, -3.-3.j])
        """
        return super().closest_symbols(received)

    def posteriors(
        self,
        received: npt.ArrayLike,
        snr: float,
        priors: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> const = komm.QAMConstellation(16)
            >>> const.posteriors([-3.1 + 2.9j], snr=2.0).round(3)
            array([0.   , 0.021, 0.219, 0.449, 0.   , 0.009, 0.091, 0.186, 0.   ,
                   0.001, 0.008, 0.016, 0.   , 0.   , 0.   , 0.   ])
        """
        return super().posteriors(received, snr, priors)



from collections import Counter
from itertools import product

import numpy as np
import numpy.typing as npt

Word = tuple[int, ...]


def is_prefix_of(w1: Word, w2: Word) -> bool:
    return len(w1) <= len(w2) and w2[: len(w1)] == w1


def is_prefix_free(words: list[Word]) -> bool:
    words = [w for w in words if len(w) > 0]  # Ignore empty words
    words = sorted(words, key=len)
    for i, w1 in enumerate(words):
        for w2 in words[i + 1 :]:
            if is_prefix_of(w1, w2):
                return False
    return True


def is_uniquely_parsable(words: list[Word]) -> bool:
    # Sardinas–Patterson algorithm. See [Say06, Sec. 2.4.1].
    augmented_words = set(words)
    while True:
        dangling_suffixes: set[Word] = set()
        for w1, w2 in product(set(words), augmented_words):
            if w1 == w2:
                continue
            if is_prefix_of(w1, w2):
                dangling_suffixes.add(w2[len(w1) :])
            elif is_prefix_of(w2, w1):
                dangling_suffixes.add(w1[len(w2) :])
        if dangling_suffixes & set(words):
            return False
        if dangling_suffixes <= augmented_words:
            return True
        augmented_words |= dangling_suffixes


def is_fully_covering(words: list[Word], cardinality: int) -> bool:
    class Node:
        def __init__(self):
            self.is_end: bool = False
            self.children: dict[int, "Node"] = {}

    # Build trie
    root = Node()
    for word in words:
        node = root
        for symbol in word:
            if symbol not in node.children:
                node.children[symbol] = Node()
            node = node.children[symbol]
        node.is_end = True

    visited = {root}
    stack = [root]
    while stack:
        node = stack.pop()
        for symbol in range(cardinality):
            if symbol not in node.children:
                return False
            child = node.children[symbol]
            if not child.is_end and child not in visited:
                visited.add(child)
                stack.append(child)

    return True


def parse_fixed_length(
    input: npt.NDArray[np.integer],
    dictionary: dict[Word, Word],
    block_size: int,
) -> npt.NDArray[np.integer]:
    if input.size % block_size != 0:
        raise ValueError(
            "length of input must be a multiple of block size"
            f" {block_size} (got {len(input)})"
        )
    try:
        output_list: list[int] = []
        for i in range(0, len(input), block_size):
            key = tuple(map(int, input[i : i + block_size]))
            output_list.extend(dictionary[key])
        output = np.asarray(output_list)
    except KeyError:
        raise ValueError("input contains invalid word")
    return output


def parse_prefix_free(
    input: npt.NDArray[np.integer],
    dictionary: dict[Word, Word],
    allow_incomplete: bool,
) -> npt.NDArray[np.integer]:
    output: list[int] = []
    i = 0
    for j in range(len(input)):
        key = tuple(map(int, input[i : j + 1]))
        if key in dictionary:
            output.extend(dictionary[key])
            i = j + 1

    if i == len(input):
        return np.asarray(output)
    elif not allow_incomplete:
        raise ValueError("input contains invalid word")

    remainder = tuple(map(int, input[i:]))
    for key, value in dictionary.items():
        if is_prefix_of(remainder, key):
            output.extend(value)
            return np.asarray(output)

    raise ValueError("input contains invalid word")


def lexicographical_code(lengths: npt.ArrayLike) -> list[Word]:
    r"""
    Generates the lexicographical prefix-free symbol code based on the given lengths.

    Parameters:
        lengths: A list where the index is the symbol and the value is its codeword length.

    Returns:
        codewords: A list where the index is the symbol and the value is the bit tuple for that symbol. Symbols with zero length receive an empty tuple.
    """
    lengths = np.asarray(lengths, dtype=int)

    if not lengths.ndim == 1:
        raise ValueError("'lengths' must be a 1D-array")
    if not np.all(lengths >= 0):
        raise ValueError("'lengths' must be non-negative")

    l_max = lengths.max()
    counts = Counter(lengths)
    integers = [0] * (l_max + 1)
    for l in range(1, l_max + 1):
        integers[l] = (integers[l - 1] + counts[l - 1]) * 2

    codewords: list[Word] = [()] * lengths.size
    for x, l in enumerate(lengths):
        codewords[x] = integer_to_symbols(integers[l], base=2, width=l)
        integers[l] += 1

    return codewords


def integer_to_symbols(integer: int, base: int, width: int) -> Word:
    symbols: list[int] = []
    for _ in range(width):
        integer, symbol = divmod(integer, base)
        symbols.append(symbol)
    return tuple(symbols[::-1])


def symbols_to_integer(symbols: npt.ArrayLike, base: int) -> int:
    symbols = np.asarray(symbols)
    integer = 0
    for symbol in symbols:
        integer = integer * base + symbol
    return integer



from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from .. import abc
from .._util import global_rng
from .._util.information_theory import arimoto_blahut, mutual_information
from .._util.validators import validate_transition_matrix


class DiscreteMemorylessChannel(abc.DiscreteMemorylessChannel):
    r"""
    General discrete memoryless channel (DMC). It is defined by an *input alphabet* $\mathcal{X}$, an *output alphabet* $\mathcal{Y}$, and a *transition probability matrix* $p_{Y \mid X}$. Here, for simplicity, the input and output alphabets are always taken as $\mathcal{X} = [0 : |\mathcal{X}|)$ and $\mathcal{Y} = [0 : |\mathcal{Y}|)$, respectively. The transition probability matrix $p_{Y \mid X}$, of size $|\mathcal{X}|$-by-$|\mathcal{Y}|$, gives the conditional probability of receiving $Y = y$ given that $X = x$ is transmitted. For more details, see <cite>CT06, Ch. 7</cite>.

    Parameters:
        transition_matrix: The channel transition probability matrix $p_{Y \mid X}$. The element in row $x \in \mathcal{X}$ and column $y \in \mathcal{Y}$ must be equal to $p_{Y \mid X}(y \mid x)$.
    """

    def __init__(
        self,
        transition_matrix: npt.ArrayLike,
        rng: np.random.Generator | None = None,
    ):
        self._transition_matrix = validate_transition_matrix(transition_matrix)
        self.rng = rng or global_rng.get()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.transition_matrix.tolist()})"

    @cached_property
    def input_cardinality(self) -> int:
        return self.transition_matrix.shape[0]

    @cached_property
    def output_cardinality(self) -> int:
        return self.transition_matrix.shape[1]

    @cached_property
    def transition_matrix(self) -> npt.NDArray[np.floating]:
        return self._transition_matrix

    def mutual_information(
        self,
        input_pmf: npt.ArrayLike,
        base: float | Literal["e"] = 2.0,
    ) -> float:
        r"""
        The mutual information is given by
        $$
            \mathrm{I}(X; Y) = \mathrm{H}(X) - \mathrm{H}(X \mid Y),
        $$
        where $\mathrm{H}(X)$ is the the entropy of $X$ and $\mathrm{H}(X \mid Y)$ is the conditional entropy of $X$ given $Y$. By default, the base of the logarithm is $2$, in which case the mutual information is measured in bits. See <cite>CT06, Ch. 2</cite>.

        Examples:
            >>> dmc = komm.DiscreteMemorylessChannel([
            ...     [0.6, 0.3, 0.1],
            ...     [0.7, 0.1, 0.2],
            ...     [0.5, 0.05, 0.45],
            ... ])
            >>> dmc.mutual_information([1/3, 1/3, 1/3])  # doctest: +FLOAT_CMP
            np.float64(0.12381109879798724)
            >>> dmc.mutual_information([1/3, 1/3, 1/3], base=3)  # doctest: +FLOAT_CMP
            np.float64(0.07811610605402552)
            >>> dmc.mutual_information([1/3, 1/3, 1/3], base='e')  # doctest: +FLOAT_CMP
            np.float64(0.08581931405385379)
        """
        return mutual_information(input_pmf, self.transition_matrix, base)

    def capacity(self, base: float | Literal["e"] = 2.0) -> float:
        r"""
        The channel capacity is given by
        $$
            C = \max_{p_X} \mathrm{I}(X; Y).
        $$
        This method computes the channel capacity via the Arimoto–Blahut algorithm. See <cite>CT06, Sec. 10.8</cite>.

        Examples:
            >>> dmc = komm.DiscreteMemorylessChannel([
            ...     [0.6, 0.3, 0.1],
            ...     [0.7, 0.1, 0.2],
            ...     [0.5, 0.05, 0.45],
            ... ])
            >>> dmc.capacity()  # doctest: +FLOAT_CMP
            np.float64(0.1616318609548566)
            >>> dmc.capacity(base=3)  # doctest: +FLOAT_CMP
            np.float64(0.10197835020154389)
            >>> dmc.capacity(base='e')  # doctest: +FLOAT_CMP
            np.float64(0.11203466870951606)
        """
        initial_guess = np.ones(self.input_cardinality) / self.input_cardinality
        input_pmf = arimoto_blahut(
            self.transition_matrix, initial_guess, max_iter=1000, tol=1e-12
        )
        return mutual_information(input_pmf, self.transition_matrix, base=base)

    def transmit(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> dmc = komm.DiscreteMemorylessChannel([
            ...     [0.9, 0.05, 0.05],
            ...     [0.0, 0.5, 0.5],
            ... ])
            >>> dmc.transmit([1, 1, 1, 0, 0, 0, 1, 0, 1, 0])
            array([2, 1, 2, 0, 0, 2, 2, 0, 1, 0])
        """
        input = np.asarray(input)
        output = np.empty_like(input, dtype=int)
        for index, symbol in np.ndenumerate(input):
            output[index] = self.rng.choice(
                a=self.output_cardinality, p=self.transition_matrix[symbol]
            )
        return output



import numpy as np
import numpy.typing as npt

from .._util.decorators import blockwise, vectorize
from .._util.docs import mkdocstrings
from .BlockCode import BlockCode


@mkdocstrings(filters=["!.*"])
class PolarCode(BlockCode):
    r"""
    Polar (Arıkan) code. Let $\mu \geq 1$ be an integer, and $\mathcal{F}$ (called the _frozen bit indices_) be a subset of $[0 : 2^\mu)$. Define $\mathcal{A} = [0 : 2^\mu) \setminus \mathcal{F}$ (called the _active bit indices_). The polar code with parameters $(\mu, \mathcal{F})$ is the [linear block code](/ref/BlockCode) whose generator matrix is obtained by selecting the rows of the order-$2^\mu$ Walsh-Hadamard matrix,
    $$ H_{2^\mu} = \begin{bmatrix} 1 & 0 \\\\ 1 & 1 \end{bmatrix} ^ {\otimes \mu}, $$
    corresponding to the active bit indices, where $\otimes$ denotes the Kronecker product. The resulting code has the following parameters:

    - Length: $n = 2^{\mu}$
    - Dimension: $k = |\mathcal{A}|$
    - Redundancy: $m = |\mathcal{F}|$

    Parameters:
        mu: The parameter $\mu$ of the code.
        frozen: The frozen bit indices $\mathcal{F}$ of the code.

    Examples:
        >>> code = komm.PolarCode(4, [0, 1, 2, 3, 4, 8])
        >>> (code.length, code.dimension, code.redundancy)
        (16, 10, 6)
        >>> code.generator_matrix
        array([[1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
               [1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
               [1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
               [1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0],
               [1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
               [1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
               [1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0],
               [1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0],
               [1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]])
        >>> code.minimum_distance()
        4
    """

    mu: int
    frozen: npt.NDArray[np.integer]

    def __init__(self, mu: int, frozen: npt.ArrayLike):
        self.mu = mu
        if not mu >= 1:
            raise ValueError("'mu' must be greater than or equal to 1")
        self.frozen = np.sort(frozen).astype(int)
        if not np.all((0 <= self.frozen) & (self.frozen < 2**mu)):
            raise ValueError("frozen bits must be between 0 and 2^mu - 1")
        if self.frozen.size != np.unique(self.frozen).size:
            raise ValueError("frozen bits must be unique")
        self.active = np.setdiff1d(np.arange(2**mu, dtype=int), self.frozen)
        hadamard = np.array([[1]])
        for _ in range(mu):
            hadamard = np.kron(hadamard, [[1, 0], [1, 1]]).astype(int)
        super().__init__(generator_matrix=hadamard[self.active])

    def __repr__(self) -> str:
        args = f"mu={self.mu}, frozen={self.frozen.tolist()}"
        return f"{self.__class__.__name__}({args})"

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        @blockwise(self.dimension)
        @vectorize
        def encode(u: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            n = 2**self.mu
            v = np.zeros((n,), dtype=int)
            v[self.active] = u
            for i in range(self.mu):
                m = 2**i
                for i in range(0, n, 2 * m):
                    v[i : i + m] ^= v[i + m : i + 2 * m]
            return v

        return encode(input)



from abc import ABC, abstractmethod
from functools import cached_property

import numpy as np
import numpy.typing as npt


class Pulse(ABC):
    @abstractmethod
    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        The waveform $p(t)$ of the pulse.
        """
        raise NotImplementedError

    @abstractmethod
    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        The spectrum $\hat{p}(f)$ of the pulse.
        """
        raise NotImplementedError

    @abstractmethod
    def autocorrelation(self, tau: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        The autocorrelation function $R(\tau)$ of the pulse.
        """
        raise NotImplementedError

    @abstractmethod
    def energy_density_spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        The energy density spectrum $S(f)$ of the pulse.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def support(self) -> tuple[float, float]:
        r"""
        The support of the pulse waveform $p(t)$, defined as the interval $[a, b]$ where $p(t)$ is non-zero.
        """
        raise NotImplementedError

    @abstractmethod
    def taps(
        self,
        samples_per_symbol: int,
        span: tuple[int, int] | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Returns the FIR taps of the pulse.

        Parameters:
            samples_per_symbol: The number of samples per symbol.
            span: The time span to consider for the taps. This parameter is optional for pulses with finite support (defaults to $[0, 1]$), but required for pulses with infinite support.
        """
        if span is None and (self.support[0] == -np.inf or self.support[1] == np.inf):
            raise ValueError(
                "pulses with infinite support require 'span' to be specified"
            )
        sps = samples_per_symbol
        start, end = span or (0, 1)
        t = np.arange(start * sps, end * sps + 1) / sps
        return self.waveform(t)



from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from .. import abc
from .._error_control_block.SingleParityCheckCode import SingleParityCheckCode
from .._util.decorators import blockwise, vectorize


@dataclass
class WagnerDecoder(abc.BlockDecoder[SingleParityCheckCode]):
    r"""
    Wagner decoder for [single parity-check codes](/ref/SingleParityCheckCode). For more details, see <cite>CF07, Sec. III.C</cite>.

    Parameters:
        code: The single parity-check code to be used for decoding.

    Notes:
        - Input type: `soft` (L-values or channel outputs).
        - Output type: `hard` (bits).
    """

    code: SingleParityCheckCode

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> code = komm.SingleParityCheckCode(4)
            >>> decoder = komm.WagnerDecoder(code)
            >>> decoder.decode([[1.52, -0.36, 1.56, 0.82], [-0.75,  1.20, -2.11, 1.73]])
            array([[0, 0, 0],
                   [1, 0, 1]])
        """

        @blockwise(self.code.length)
        @vectorize
        def decode(r: npt.NDArray[np.integer]):
            v_hat = (r < 0).astype(int)
            if np.count_nonzero(v_hat) % 2 != 0:
                i = np.argmin(np.abs(r))
                v_hat[i] ^= 1
            u_hat = self.code.project_word(v_hat)
            return u_hat

        return decode(input)



from collections.abc import Callable
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .. import abc


class ScalarQuantizer(abc.ScalarQuantizer):
    r"""
    General scalar quantizer. It is defined by a list of *levels*, $v_0, v_1, \ldots, v_{L-1}$, and a list of *thresholds*, $t_0, t_1, \ldots, t_L$, satisfying
    $$
        -\infty = t_0 < v_0 < t_1 < v_1 < \cdots < t_{L - 1} < v_{L - 1} < t_L = +\infty.
    $$
    Given an input $x \in \mathbb{R}$, the output of the quantizer is given by $y = v_i$ if and only if $t_i \leq x < t_{i+1}$, where $i \in [0:L)$. For more details, see <cite>Say06, Ch. 9</cite>.

    Parameters:
        levels: The quantizer levels $v_0, v_1, \ldots, v_{L-1}$. It should be a list floats of length $L$.

        thresholds: The quantizer finite thresholds $t_1, t_2, \ldots, t_{L-1}$. It should be a list of floats of length $L - 1$.

    Examples:
        The $5$-level scalar quantizer whose characteristic (input × output) curve is depicted in the figure below has levels
        $$
            v_0 = -2, ~ v_1 = -1, ~ v_2 = 0, ~ v_3 = 1, ~ v_4 = 2,
        $$
        and thresholds
        $$
            t_0 = -\infty, ~ t_1 = -1.5, ~ t_2 = -0.3, ~ t_3 = 0.8, ~ t_4 = 1.4, ~ t_5 = \infty.
        $$

        <figure markdown>
          ![Scalar quantizer example.](/fig/scalar_quantizer_5.svg)
        </figure>

        >>> quantizer = komm.ScalarQuantizer(
        ...     levels=[-2.0, -1.0, 0.0, 1.0, 2.0],
        ...     thresholds=[-1.5, -0.3, 0.8, 1.4],
        ... )
    """

    def __init__(self, levels: npt.ArrayLike, thresholds: npt.ArrayLike) -> None:
        self._levels = np.asarray(levels, dtype=float)
        self._thresholds = np.asarray(thresholds, dtype=float)
        self.__post_init__()

    def __post_init__(self) -> None:
        if not self.thresholds.size == self.levels.size - 1:
            raise ValueError("'len(thresholds)' must be equal to 'len(levels) - 1'")

        interleaved = np.empty(2 * self.levels.size - 1, dtype=float)
        interleaved[0::2] = self.levels
        interleaved[1::2] = self.thresholds

        if not np.array_equal(np.unique(interleaved), interleaved):
            raise ValueError("invalid values for 'levels' and 'thresholds'")

    def __repr__(self) -> str:
        args = ", ".join([
            f"levels={self.levels.tolist()}",
            f"thresholds={self.thresholds.tolist()}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def levels(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.ScalarQuantizer(
            ...     levels=[-2.0, -1.0, 0.0, 1.0, 2.0],
            ...     thresholds=[-1.5, -0.3, 0.8, 1.4],
            ... )
            >>> quantizer.levels
            array([-2., -1.,  0.,  1.,  2.])
        """
        return self._levels

    @cached_property
    def thresholds(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.ScalarQuantizer(
            ...     levels=[-2.0, -1.0, 0.0, 1.0, 2.0],
            ...     thresholds=[-1.5, -0.3, 0.8, 1.4],
            ... )
            >>> quantizer.thresholds
            array([-1.5, -0.3,  0.8,  1.4])
        """
        return self._thresholds

    def mean_squared_error(
        self,
        input_pdf: Callable[[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        input_range: tuple[float, float],
        points_per_interval: int = 4096,
    ) -> float:
        r"""
        Examples:
            >>> quantizer = komm.ScalarQuantizer(
            ...     levels=[-2.0, -1.0, 0.0, 1.0, 2.0],
            ...     thresholds=[-1.5, -0.3, 0.8, 1.4],
            ... )
            >>> gaussian_pdf = lambda x: 1/np.sqrt(2*np.pi) * np.exp(-x**2/2)
            >>> quantizer.mean_squared_error(
            ...     input_pdf=gaussian_pdf,
            ...     input_range=(-5, 5),
            ... )  # doctest: +FLOAT_CMP
            0.13598089455499335
        """
        return super().mean_squared_error(input_pdf, input_range, points_per_interval)

    def digitize(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> quantizer = komm.ScalarQuantizer(
            ...     levels=[-2.0, -1.0, 0.0, 1.0, 2.0],
            ...     thresholds=[-1.5, -0.3, 0.8, 1.4],
            ... )
            >>> quantizer.digitize([-2.5, -2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2, 2.5])
            array([0, 0, 1, 1, 1, 2, 2, 3, 4, 4, 4])
        """
        return super().digitize(input)

    def quantize(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.ScalarQuantizer(
            ...     levels=[-2.0, -1.0, 0.0, 1.0, 2.0],
            ...     thresholds=[-1.5, -0.3, 0.8, 1.4],
            ... )
            >>> quantizer.quantize([-2.5, -2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2, 2.5])
            array([-2., -2., -1., -1., -1.,  0.,  0.,  1.,  2.,  2.,  2.])
        """
        return super().quantize(input)



from functools import cache, reduce
from itertools import combinations

import numpy as np
import numpy.typing as npt
from typeguard import typechecked

from .._util.bit_operations import int_to_bits
from .._util.docs import mkdocstrings
from .BlockCode import BlockCode


@mkdocstrings(members=["reed_partitions"], filters=["!.*"])
@typechecked
class ReedMullerCode(BlockCode):
    r"""
    Reed–Muller code. Let $\mu$ and $\rho$ be two integers such that $0 \leq \rho < \mu$. The Reed–Muller code with parameters $(\rho, \mu)$ is the [linear block code](/ref/BlockCode) whose generator matrix rows are
    $$
        \mathbf{1}, \, v_1, \ldots, v_m, \, v_1 v_2, v_1 v_3, \ldots, v_{\mu} v_{\mu - 1}, \, \ldots \, \text{(up to products of degree $\rho$)},
    $$
    where $\mathbf{1}$ is the all-one $2^\mu$-tuple, $v_i$ is the binary $2^\mu$-tuple composed of $2^{\mu-i}$ repetitions of alternating blocks of zeros and ones, each block of length $2^i$, for $1 \leq i \leq \mu$, and $v_i v_j$ denotes the component-wise product (logical AND) of $v_i$ and $v_j$. Here we take the rows in reverse order. The resulting code has the following parameters:

    - Length: $n = 2^{\mu}$
    - Dimension: $k = 1 + {\mu \choose 1} + \cdots + {\mu \choose \rho}$
    - Redundancy: $m = 1 + {\mu \choose 1} + \cdots + {\mu \choose \mu - \rho - 1}$
    - Minimum distance: $d = 2^{\mu - \rho}$

    For more details, see <cite>LC04, Sec. 4.3</cite>.

    Notes:
        - For $\rho = 0$ it reduces to a [repetition code](/ref/RepetitionCode).
        - For $\rho = 1$ it reduces to a lengthened [simplex code](/ref/SimplexCode).
        - For $\rho = \mu - 2$ it reduces to an extended [Hamming code](/ref/HammingCode).
        - For $\rho = \mu - 1$ it reduces to a [single parity-check code](/ref/SingleParityCheckCode).

    Parameters:
        rho: The parameter $\rho$ of the code.
        mu: The parameter $\mu$ of the code.

    Examples:
        >>> code = komm.ReedMullerCode(2, 4)
        >>> (code.length, code.dimension, code.redundancy)
        (16, 11, 5)
        >>> code.generator_matrix
        array([[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1],
               [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1],
               [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1],
               [0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1],
               [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1],
               [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
               [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1],
               [0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
               [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
               [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]])
        >>> code.minimum_distance()
        4
    """

    def __init__(self, rho: int, mu: int) -> None:
        if not 0 <= rho < mu:
            raise ValueError("'rho' and 'mu' must satisfy 0 <= rho < mu")
        self.rho = rho
        self.mu = mu
        super().__init__(generator_matrix=reed_muller_generator_matrix(rho, mu))

    def __repr__(self) -> str:
        args = f"rho={self.rho}, mu={self.mu}"
        return f"{self.__class__.__name__}({args})"

    @cache
    def minimum_distance(self) -> int:
        return 2 ** (self.mu - self.rho)

    @cache
    def reed_partitions(self) -> list[npt.NDArray[np.integer]]:
        r"""
        The Reed partitions of the code. See <cite>LC04, Sec. 4.3</cite>.

        Examples:
            >>> code = komm.ReedMullerCode(2, 4)
            >>> reed_partitions = code.reed_partitions()
            >>> reed_partitions[1]
            array([[ 0,  1,  4,  5],
                   [ 2,  3,  6,  7],
                   [ 8,  9, 12, 13],
                   [10, 11, 14, 15]])
            >>> reed_partitions[8]
            array([[ 0,  4],
                   [ 1,  5],
                   [ 2,  6],
                   [ 3,  7],
                   [ 8, 12],
                   [ 9, 13],
                   [10, 14],
                   [11, 15]])
        """
        rho, mu = self.rho, self.mu
        reed_partitions: list[npt.NDArray[np.integer]] = []
        bin_vectors = [int_to_bits(range(2**ell), width=ell) for ell in range(mu + 1)]
        for ell in range(rho, -1, -1):
            for indices in combinations(range(mu), ell):
                setI = np.array(indices, dtype=int)
                setE = np.setdiff1d(np.arange(mu), indices, assume_unique=True)
                setS = np.dot(bin_vectors[ell], 2**setI)
                setQ = np.dot(bin_vectors[mu - ell], 2**setE)
                reed_partitions.append(setS[np.newaxis] + setQ[np.newaxis].T)
        return reed_partitions


def reed_muller_generator_matrix(rho: int, mu: int) -> npt.NDArray[np.integer]:
    # See [LC04, p. 105–114]. Assumes 0 <= rho < mu.
    v = np.empty((mu, 2**mu), dtype=int)
    for i in range(mu):
        block = np.hstack((
            np.zeros(2 ** (mu - i - 1), dtype=int),
            np.ones(2 ** (mu - i - 1), dtype=int),
        ))
        v[mu - i - 1] = np.tile(block, 2**i)

    G_list: list[npt.NDArray[np.integer]] = []
    for ell in range(rho, 0, -1):
        for indices in combinations(range(mu), ell):
            row = reduce(np.multiply, v[indices, :])
            G_list.append(row)
    row = np.ones(2**mu, dtype=int)
    G_list.append(row)

    return np.array(G_list, dtype=int)



from typing import Literal

from .._util.bit_operations import bits_to_int, int_to_bits
from .BinarySequence import BinarySequence
from .sequences import hadamard_matrix


class WalshHadamardSequence(BinarySequence):
    r"""
    Walsh–Hadamard sequence. Consider the following recursive matrix construction:
    $$
        H_1 =
        \begin{bmatrix}
            +1
        \end{bmatrix}, \qquad
        H_{2^n} =
        \begin{bmatrix}
            H_{2^{n-1}} & H_{2^{n-1}} \\\\
            H_{2^{n-1}} & -H_{2^{n-1}}
        \end{bmatrix},
    $$
    for $n = 1, 2, \ldots$. For example, for $n = 3$,
    $$
        H_8 =
        \begin{bmatrix}
            +1 & +1 & +1 & +1 & +1 & +1 & +1 & +1 \\\\
            +1 & -1 & +1 & -1 & +1 & -1 & +1 & -1 \\\\
            +1 & +1 & -1 & -1 & +1 & +1 & -1 & -1 \\\\
            +1 & -1 & -1 & +1 & +1 & -1 & -1 & +1 \\\\
            +1 & +1 & +1 & +1 & -1 & -1 & -1 & -1 \\\\
            +1 & -1 & +1 & -1 & -1 & +1 & -1 & +1 \\\\
            +1 & +1 & -1 & -1 & -1 & -1 & +1 & +1 \\\\
            +1 & -1 & -1 & +1 & -1 & +1 & +1 & -1 \\\\
        \end{bmatrix}
    $$
    The above matrix is said to be in *natural ordering*. If the rows of the matrix are rearranged by first applying the bit-reversal permutation and then the Gray-code permutation, the following matrix is obtained:
    $$
        H_8^{\mathrm{s}} =
        \begin{bmatrix}
            +1 & +1 & +1 & +1 & +1 & +1 & +1 & +1 \\\\
            +1 & +1 & +1 & +1 & -1 & -1 & -1 & -1 \\\\
            +1 & +1 & -1 & -1 & -1 & -1 & +1 & +1 \\\\
            +1 & +1 & -1 & -1 & +1 & +1 & -1 & -1 \\\\
            +1 & -1 & -1 & +1 & +1 & -1 & -1 & +1 \\\\
            +1 & -1 & -1 & +1 & -1 & +1 & +1 & -1 \\\\
            +1 & -1 & +1 & -1 & -1 & +1 & -1 & +1 \\\\
            +1 & -1 & +1 & -1 & +1 & -1 & +1 & -1 \\\\
        \end{bmatrix}
    $$
    The above matrix is said to be in *sequency ordering*. It has the property that row $i$ has exactly $i$ sign changes.

    The Walsh–Hadamard sequence of *length* $L$ and *index* $i \in [0 : L)$ is a [binary sequence](/ref/BinarySequence) whose polar format is the $i$-th row of $H_L$, if assuming natural ordering, or $H_L^{\mathrm{s}}$, if assuming sequency ordering. Fore more details, see [Wikipedia: Hadamard matrix](https://en.wikipedia.org/wiki/Hadamard_matrix) and [Wikipedia: Walsh matrix](https://en.wikipedia.org/wiki/Walsh_matrix).

    Parameters:
        length: Length $L$ of the Walsh–Hadamard sequence. Must be a power of two.

        ordering: Ordering to be assumed. Should be one of `'natural'`, `'sequency'`, or `'dyadic'`. The default value is `'natural'`.

        index: Index of the Walsh–Hadamard sequence, with respect to the ordering assumed. Must be in the set $[0 : L)$. The default value is `0`.

    Examples:
        >>> walsh_hadamard = komm.WalshHadamardSequence(length=8, ordering='natural', index=5)
        >>> walsh_hadamard.polar_sequence
        array([ 1, -1,  1, -1, -1,  1, -1,  1])


        >>> walsh_hadamard = komm.WalshHadamardSequence(length=8, ordering='sequency', index=5)
        >>> walsh_hadamard.polar_sequence
        array([ 1, -1, -1,  1, -1,  1,  1, -1])

        >>> walsh_hadamard = komm.WalshHadamardSequence(length=8, ordering='dyadic', index=5)
        Traceback (most recent call last):
        ...
        NotImplementedError
    """

    def __init__(
        self,
        length: int,
        ordering: Literal["natural", "sequency", "dyadic"] = "natural",
        index: int = 0,
    ) -> None:
        if length & (length - 1):
            raise ValueError("'length' must be a power of two")
        if not 0 <= index < length:
            raise ValueError("'index' must be in [0 : length)")
        if ordering not in {"natural", "sequency", "dyadic"}:
            raise ValueError("'ordering' must be in {'natural', 'sequency', 'dyadic'}")

        if ordering == "natural":
            natural_index = index
        elif ordering == "sequency":
            index_gray = index ^ (index >> 1)
            width = (length - 1).bit_length()
            natural_index = bits_to_int(int_to_bits(index_gray, width, "MSB-first"))
        elif ordering == "dyadic":
            raise NotImplementedError

        self.index = index
        self.ordering = ordering
        super().__init__(polar_sequence=hadamard_matrix(length)[natural_index])

    def __repr__(self) -> str:
        args = f"length={self.length}, ordering={self.ordering!r}, index={self.index}"
        return f"{self.__class__.__name__}({args})"



import numpy as np
import numpy.typing as npt


def rect(x: npt.ArrayLike) -> npt.NDArray[np.floating]:
    x = np.asarray(x)
    return 1.0 * (-0.5 <= x) * (x < 0.5)


def tri(x: npt.ArrayLike) -> npt.NDArray[np.floating]:
    x = np.asarray(x)
    return (1.0 - np.abs(x)) * (-1.0 <= x) * (x < 1.0) + 0.0


def raised_cosine(x: npt.ArrayLike, α: float) -> npt.NDArray[np.floating]:
    x = np.asarray(x)
    return np.pi / 4 * np.sinc(x) * (np.sinc(α * x + 0.5) + np.sinc(α * x - 0.5))



from functools import cache

import numpy as np
import numpy.typing as npt
from tqdm import tqdm
from typeguard import typechecked

from .._util.bit_operations import int_to_bits
from .._util.docs import mkdocstrings
from .BlockCode import BlockCode


@mkdocstrings(filters=["!.*"])
@typechecked
class Lexicode(BlockCode):
    r"""
    Lexicographic code (lexicode). For a given length $n$ and minimum distance $d$, it is the [linear block code](/ref/BlockCode) obtained by starting with the all-zero codeword and adding all binary $n$-tuples (in lexicographic order) that are at least at distance $d$ from all codewords already in the code.

    Parameters:
        n: The length $n$ of the code.
        d: The minimum distance $d$ of the code.

    For more details, see <cite>HP03, Sec. 2.11</cite>.

    Examples:
        >>> code = komm.Lexicode(7, 3)  # Hamming (7, 4)
        >>> (code.length, code.dimension, code.redundancy)
        (7, 4, 3)
        >>> code.generator_matrix
        array([[0, 0, 0, 0, 1, 1, 1],
               [0, 0, 1, 1, 0, 0, 1],
               [0, 1, 0, 1, 0, 1, 0],
               [1, 0, 0, 1, 0, 1, 1]])
        >>> code.minimum_distance()
        3
    """

    def __init__(self, n: int, d: int) -> None:
        if not 1 <= d <= n:
            raise ValueError("'n' and 'd' must satisfy 1 <= d <= n")
        self.n = n
        self.d = d
        super().__init__(generator_matrix=lexicode_generator_matrix(n, d))

    def __repr__(self) -> str:
        args = f"n={self.n}, d={self.d}"
        return f"{self.__class__.__name__}({args})"

    @cache
    def minimum_distance(self) -> int:
        return self.d


def lexicode_generator_matrix(n: int, d: int) -> npt.NDArray[np.integer]:
    codewords = [0]
    basis: list[int] = []
    for i in tqdm(range(1, 2**n), desc="Generating lexicode", delay=2.5):
        # Reverse checking is way faster [Wikipedia].
        if all((c ^ i).bit_count() >= d for c in reversed(codewords)):
            if len(codewords).bit_count() == 1:  # Is power of 2.
                basis.append(i)
            codewords.append(i)
    generator_matrix = int_to_bits(basis, width=n, bit_order="MSB-first")
    return generator_matrix



from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from .._util import global_rng
from .._util.information_theory import entropy
from .._util.validators import validate_pmf
from ..types import Array1D


class DiscreteMemorylessSource:
    r"""
    Discrete memoryless source. It is defined by a finite *source alphabet* $\mathcal{X}$ and a *pmf* $p$ over $\mathcal{X}$. The value of $p(x)$ gives the probability of the source emitting symbol $x \in \mathcal{X}$, with the probability of emitting a symbol being independent of all previously emitted symbols. Here, for simplicity, the alphabet is taken as $\mathcal{X} = [0 : |\mathcal{X}|)$, where $|\mathcal{X}|$ is called the *source cardinality*.

    Parameters:
        pmf: Either a one-dimensional array of floats representing the source pmf $p$, or a positive integer $M$, in which case a uniform pmf over $[0 : M)$ is assumed.

    Note:
        The cardinality $|\mathcal{X}|$ is inferred from the length of the `pmf` array.
    """

    def __init__(
        self,
        pmf: npt.ArrayLike | int,
        rng: np.random.Generator | None = None,
    ):
        if isinstance(pmf, int):
            if not pmf >= 1:
                raise ValueError("cardinality must be at least 1")
            pmf = np.full(pmf, 1 / pmf)
        self._pmf = validate_pmf(pmf, joint=False)
        self._rng = rng or global_rng.get()

    def __repr__(self) -> str:
        return f"{__class__.__name__}(pmf={self.pmf.tolist()})"

    @cached_property
    def pmf(self) -> Array1D[np.floating]:
        r"""
        The source pmf $p$ over $\mathcal{X}$.

        Examples:
            >>> source = komm.DiscreteMemorylessSource([1/2, 1/4, 1/8, 1/8])
            >>> source.pmf
            array([0.5  , 0.25 , 0.125, 0.125])

            >>> source = komm.DiscreteMemorylessSource(4)
            >>> source.pmf
            array([0.25, 0.25, 0.25, 0.25])
        """
        return self._pmf

    @cached_property
    def cardinality(self) -> int:
        r"""
        The source cardinality $|\mathcal{X}|$.

        Examples:
            >>> source = komm.DiscreteMemorylessSource([1/2, 1/4, 1/8, 1/8])
            >>> source.cardinality
            4

            >>> source = komm.DiscreteMemorylessSource(4)
            >>> source.cardinality
            4
        """
        return self.pmf.size

    def entropy_rate(self, base: float | Literal["e"] = 2.0) -> np.floating:
        r"""
        Computes the source entropy rate. For a discrete memoryless source, this is simply the [entropy](/ref/entropy) of the pmf $p$.

        Parameters:
            base: See [`komm.entropy`](/ref/entropy). The default value is `2.0`.

        Returns:
            entropy_rate: The source entropy rate.

        Examples:
            >>> source = komm.DiscreteMemorylessSource([1/2, 1/4, 1/8, 1/8])
            >>> source.entropy_rate()
            np.float64(1.75)
            >>> source.entropy_rate(base=4)
            np.float64(0.875)

            >>> source = komm.DiscreteMemorylessSource(4)
            >>> source.entropy_rate()
            np.float64(2.0)
            >>> source.entropy_rate(base=4)
            np.float64(1.0)
        """
        return entropy(self.pmf, base)

    def emit(
        self,
        shape: tuple[int, ...] | int | None = None,
    ) -> npt.NDArray[np.integer]:
        r"""
        Returns random symbols from the source.

        Parameters:
            shape: The shape of the output array. The default value corresponds to a single symbol.

        Returns:
            symbols: The emitted symbols from the source. It is an array with elements in $\mathcal{X}$ and the given shape.

        Examples:
            >>> source = komm.DiscreteMemorylessSource([1/2, 1/4, 1/8, 1/8])
            >>> source.emit()
            array([2])
            >>> source.emit(10)
            array([0, 2, 1, 0, 3, 2, 2, 0, 0, 0])
            >>> source.emit((2, 5))
            array([[3, 1, 2, 0, 0],
                   [1, 0, 2, 1, 2]])
        """
        if shape is None:
            shape = (1,)
        elif isinstance(shape, int):
            shape = (shape,)
        return self._rng.choice(self.pmf.size, p=self.pmf, size=shape)



from abc import ABC, abstractmethod
from functools import cache, cached_property, reduce
from heapq import heappop, heappush
from itertools import combinations

import numpy as np
import numpy.typing as npt

from .._algebra import domain, ring
from .._algebra.BinaryPolynomial import BinaryPolynomial
from .._algebra.BinaryPolynomialFraction import BinaryPolynomialFraction
from .._finite_state_machine.MealyMachine import MealyMachine
from .._util.bit_operations import bits_to_int, int_to_bits
from .._util.matrices import invariant_factors
from ..types import Array1D, Array2D


class ConvolutionalCode(ABC):
    @cached_property
    @abstractmethod
    def num_input_bits(self) -> int:
        r"""
        The number of input bits per block, $k$.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def num_output_bits(self) -> int:
        r"""
        The number of output bits per block, $n$.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def degree(self) -> int:
        r"""
        The *degree* $\sigma$ of the encoder. This corresponds to the number of delay elements in the encoder realization.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def generator_matrix(self) -> Array2D[np.object_]:
        r"""
        Returns the (transform-domain) *generator matrix* (also known as *transfer function matrix*) $G(D)$ of the encoder. This is a $k \times n$ array of [binary polynomial fractions](/ref/BinaryPolynomialFraction).
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def constraint_lengths(self) -> Array1D[np.integer]:
        r"""
        The *constraint lengths* $\nu_i$ of the encoder, for $i \in [0 : k)$. These are defined by
        $$
            \nu_i = \max \\{ \deg p_{i,0}(D), \deg p_{i,1}(D), \ldots, \deg p_{i,n-1}(D), \deg q_i(D) \\},
        $$
        where $p_{i,j}(D) / q_i(D)$ are the entries of the $i$-th row of the generator matrix $G(D)$, satisfying
        $$
            \gcd(p_{i,0}(D), p_{i,1}(D), \ldots, p_{i,n-1}(D), q_i(D)) = 1.
        $$
        This is a $k$-array of integers.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def overall_constraint_length(self) -> int:
        r"""
        The *overall constraint length* $\nu$ of the encoder, defined by
        $$
            \nu = \sum_{i \in [0:k)} \nu_i,
        $$
        where $\nu_i$ are the constraint lengths of the encoder.
        """
        return int(np.sum(self.constraint_lengths))

    @cached_property
    @abstractmethod
    def memory_order(self) -> int:
        r"""
        The *memory order* $\mu$ of the encoder. It is given by
        $$
            \mu = \max_{i \in [0:k)} \nu_i,
        $$
        where $\nu_i$ are the constraint lengths of the encoder.
        """
        return int(np.max(self.constraint_lengths))

    @cache
    @abstractmethod
    def state_space_representation(
        self,
    ) -> tuple[
        Array2D[np.integer],
        Array2D[np.integer],
        Array2D[np.integer],
        Array2D[np.integer],
    ]:
        r"""
        Returns the matrices $(\mathbf{A}, \mathbf{B}, \mathbf{C}, \mathbf{D})$ corresponding to the state-space representation of the encoder realization. The *state-space representation* of the encoder is given by
        $$
        \begin{aligned}
            s_{t+1} & = s_t \mathbf{A} + u_t \mathbf{B}, \\\\
            v_t & = s_t \mathbf{C} + u_t \mathbf{D},
        \end{aligned}
        $$
        where

        - $u_t \in \mathbb{B}^k$ is the *input block*,
        - $v_t \in \mathbb{B}^n$ is the *output block*,
        - $s_t \in \mathbb{B}^\sigma$ is the *state*,
        - $\mathbf{A} \in \mathbb{B}^{\sigma \times \sigma}$ is the *state matrix*,
        - $\mathbf{B} \in \mathbb{B}^{k \times \sigma}$ is the *control matrix*,
        - $\mathbf{C} \in \mathbb{B}^{\sigma \times n}$ is the *observation matrix*,
        - $\mathbf{D} \in \mathbb{B}^{k \times n}$ is the *transition matrix*.

        Returns:
            state_matrix: The state matrix $\mathbf{A}$ of the encoder.
            control_matrix: The control matrix $\mathbf{B}$ of the encoder.
            observation_matrix: The observation matrix $\mathbf{C}$ of the encoder.
            transition_matrix: The transition matrix $\mathbf{D}$ of the encoder.
        """
        raise NotImplementedError

    @cache
    @abstractmethod
    def finite_state_machine(self) -> MealyMachine:
        r"""
        Returns the [finite-state (Mealy) machine](/ref/MealyMachine) of the encoder.
        """
        k, σ = self.num_input_bits, self.degree
        transitions = np.empty((2**σ, 2**k), dtype=int)
        outputs = np.empty((2**σ, 2**k), dtype=int)
        for s, x in np.ndindex(2**σ, 2**k):
            initial_state = int_to_bits(s, width=σ)
            u = int_to_bits(x, width=k)
            v, final_state = self.encode_with_state(u, initial_state)
            transitions[s, x] = bits_to_int(final_state)
            outputs[s, x] = bits_to_int(v)
        return MealyMachine(transitions, outputs)

    @cache
    def _minors_gcd(self) -> BinaryPolynomial:
        # We keep this method for testing purposes.
        G_mat = self.generator_matrix
        k, n = G_mat.shape
        if any(x.denominator != 0b1 for row in G_mat for x in row):
            raise ValueError("this method only works with polynomial matrices")
        G_poly: list[list[BinaryPolynomial]] = [
            [x.numerator for x in row] for row in G_mat
        ]
        minors: list[BinaryPolynomial] = []
        for js in combinations(range(n), k):
            submatrix = [[row[j] for j in js] for row in G_poly]
            minors.append(ring.determinant(submatrix))
        return reduce(domain.gcd, minors)

    @cache
    @abstractmethod
    def is_catastrophic(self) -> bool:
        """
        Returns whether the encoder is catastrophic. A convolutional encoder is *catastrophic* if there exists an infinite-weight input sequence that generates a finite-weight output sequence.
        """
        # See [JZ15, Secs. 2.2 and 2.3], in particular Theorem 2.10.
        G_mat = self.generator_matrix
        denominators = [x.denominator for row in G_mat for x in row]
        q = reduce(domain.lcm, denominators)
        G_prime = np.array([
            [(BinaryPolynomialFraction(q) * x).numerator for x in row] for row in G_mat
        ])
        factors = invariant_factors(G_prime)
        if len(factors) < G_mat.shape[0]:
            raise ValueError("generator matrix is not full rank")
        ɑ_k = BinaryPolynomialFraction(factors[-1], q).numerator
        return len(ɑ_k.exponents()) != 1

    @cache
    @abstractmethod
    def free_distance(self) -> int:
        r"""
        Returns the *free distance* $d_\mathrm{free}$ of the code. This is equal to the minimum Hamming weight among all possible non-zero output sequences.
        """
        fsm = self.finite_state_machine()
        heap = [(0, 0)]  # (weight, state)
        best = [np.inf] * fsm.num_states
        while heap:
            w, s = heappop(heap)
            if s == 0 and w > 0:
                return int(w)
            for x in range(fsm.num_input_symbols):
                s1 = fsm.transitions[s][x]
                w1 = w + fsm.outputs[s][x].bit_count()
                if w1 < best[s1]:
                    best[s1] = w1
                    heappush(heap, (w1, s1))
                elif s1 == 0 and w1 > 0:
                    heappush(heap, (w1, 0))
        raise RuntimeError

    @abstractmethod
    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Encodes a given bit sequence, starting from the all-zero state.

        Parameters:
            input: The bit sequence to be encoded. Must be a 1D-array of bits, with length multiple of $k$.

        Returns:
            output: The encoded bit sequence. It is a 1D-array of bits, with length multiple of $n$.
        """
        σ = self.degree
        output, _ = self.encode_with_state(input, np.zeros(σ, dtype=int))
        return output

    @abstractmethod
    def encode_with_state(
        self,
        input: npt.ArrayLike,
        initial_state: npt.ArrayLike,
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
        r"""
        Encodes a given bit sequence, starting from a given state.

        Parameters:
            input: The bit sequence to be encoded. Must be a 1D-array of bits, with length multiple of $k$.
            initial_state: The initial state. Must be a 1D-array of length $\sigma$.

        Returns:
            output: The encoded bit sequence. It is a 1D-array of bits, with length multiple of $n$.
            final_state: The final state. It is a 1D-array of length $\sigma$.
        """
        n, k, σ = self.num_output_bits, self.num_input_bits, self.degree
        A_mat, B_mat, C_mat, D_mat = self.state_space_representation()

        input = np.asarray(input).reshape((-1, k))
        state = np.asarray(initial_state)

        if state.ndim != 1:
            raise ValueError("'initial_state' must be a 1D-array")
        if state.size != σ:
            raise ValueError(
                "length of 'initial_state' must be 'degree' "
                f"(expected {σ}, got {state.size})"
            )

        output = np.empty(n * input.size // k, dtype=int)
        for t, u in enumerate(input):
            output[t * n : (t + 1) * n] = (state @ C_mat + u @ D_mat) % 2
            state = (state @ A_mat + u @ B_mat) % 2

        return output, state



from functools import cache, cached_property
from itertools import product

import numpy as np
import numpy.typing as npt

from .. import abc
from .._algebra.BinaryPolynomial import BinaryPolynomial
from .._algebra.BinaryPolynomialFraction import BinaryPolynomialFraction
from .._finite_state_machine.MealyMachine import MealyMachine
from .._util.format import format_list_no_quotes as fmt
from .._util.matrices import block_diagonal
from ..types import Array1D, Array2D


class ConvolutionalCode(abc.ConvolutionalCode):
    r"""
    Binary convolutional encoder. It is characterized by a *matrix of feedforward polynomials* $P(D)$, of shape $k \times n$, and (optionally) by a *vector of feedback polynomials* $q(D)$, of length $k$. The parameters $k$ and $n$ are the number of input and output bits per block, respectively. In this class, the encoder is realized in controllable canonical form. For more details, see <cite>McE98</cite>, <cite>JZ15</cite>, and <cite>LC04, Chs. 11, 12</cite>.

    Parameters:
        feedforward_polynomials: The matrix of feedforward polynomials $P(D)$, which is a $k \times n$ matrix whose entries are either [binary polynomials](/ref/BinaryPolynomial) or integers to be converted to the former.

        feedback_polynomials: The vector of feedback polynomials $q(D)$, which is a $k$-vector whose entries are either [binary polynomials](/ref/BinaryPolynomial) or integers to be converted to the former. The default value corresponds to no feedback, that is, $q_i(D) = 1$ for all $i \in [0 : k)$.

    Examples:
        1. Consider the encoder with parameters $(n, k, \sigma) = (3, 2, 7)$ depicted below.

            <figure markdown>
            ![Convolutional encoder with parameters (3, 2, 7).](/fig/cc_3_2_7.svg)
            </figure>

            Its matrix of feedforward polynomials is given by
            $$
                P(D) =
                \begin{bmatrix}
                    D^4 + D^3 + 1  &  D^4 + D^2 + D + 1  &  0 \\\\
                    0  &  D^3 + D  &  D^3 + D^2 + 1
                \end{bmatrix}.
            $$

                >>> code = komm.ConvolutionalCode(
                ...     feedforward_polynomials=[
                ...         [0b11001, 0b10111,      0],
                ...         [      0,  0b1010, 0b1101],
                ...     ],
                ... )
                >>> code = komm.ConvolutionalCode(
                ...     feedforward_polynomials=[[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]],
                ... )

        1. Consider the feedback encoder with parameters $(n, k, \sigma) = (2, 1, 4)$ depicted below.

            <figure markdown>
            ![Feedback convolutional encoder with parameters (2, 1, 4).](/fig/cc_2_1_4_fb.svg)
            </figure>

            Its matrix of feedforward polynomials is given by
            $$
                P(D) =
                \begin{bmatrix}
                    D^4 + D^2 + D + 1 && D^4 + D^3 + 1
                \end{bmatrix},
            $$
            and its vector of feedback polynomials is given by
            $$
                q(D) =
                \begin{bmatrix}
                    D^4 + D^2 + D + 1
                \end{bmatrix}.
            $$

                >>> code = komm.ConvolutionalCode(
                ...     feedforward_polynomials=[[0b10111, 0b11001]],
                ...     feedback_polynomials=[0b10111],
                ... )
                >>> code = komm.ConvolutionalCode(
                ...     feedforward_polynomials=[[0o27, 0o31]],
                ...     feedback_polynomials=[0o27],
                ... )

    """

    feedforward_polynomials: list[list[BinaryPolynomial]]
    feedback_polynomials: list[BinaryPolynomial]

    def __init__(
        self,
        feedforward_polynomials: npt.ArrayLike,
        feedback_polynomials: npt.ArrayLike | None = None,
    ) -> None:
        ff = np.asarray(feedforward_polynomials, dtype=int)
        if ff.ndim != 2:
            raise ValueError("feedforward must be a 2-dimensional array")
        if feedback_polynomials is None:
            fb = np.ones(ff.shape[0], dtype=int)
        else:
            fb = np.asarray(feedback_polynomials, dtype=int)
        if fb.ndim != 1:
            raise ValueError("feedback must be a 1-dimensional array")
        if fb.shape[0] != ff.shape[0]:
            raise ValueError("feedback and feedforward dimensions do not match")
        self.feedforward_polynomials = [[BinaryPolynomial(p) for p in ps] for ps in ff]
        self.feedback_polynomials = [BinaryPolynomial(q) for q in fb]

    def __repr__(self) -> str:
        args = f"feedforward_polynomials={fmt(self.feedforward_polynomials)}"
        if not np.all([q == 1 for q in self.feedback_polynomials]):
            args += f", feedback_polynomials={fmt(self.feedback_polynomials)}"
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def num_input_bits(self) -> int:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]])
            >>> code.num_input_bits
            2
        """
        return len(self.feedforward_polynomials)

    @cached_property
    def num_output_bits(self) -> int:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]])
            >>> code.num_output_bits
            3
        """
        return len(self.feedforward_polynomials[0])

    @cached_property
    def degree(self) -> int:
        r"""
        For a convolutional encoder realized in controllable canonical form, the degree $\sigma$ is equal to the overall constraint length $\nu$.

        Examples:
            >>> code = komm.ConvolutionalCode([[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]])
            >>> code.degree
            7
        """
        return self.overall_constraint_length

    @cached_property
    def generator_matrix(self) -> Array2D[np.object_]:
        r"""
        For a convolutional code with matrix of feedforward polynomials
        $$
            P(D) =
            \begin{bmatrix}
                p_{0,0}(D)   & p_{0,1}(D)   & \cdots & p_{0,n-1}(D)   \\\\
                p_{1,0}(D)   & p_{1,1}(D)   & \cdots & p_{1,n-1}(D)   \\\\
                \vdots       & \vdots       & \ddots & \vdots         \\\\
                p_{k-1,0}(D) & p_{k-1,1}(D) & \cdots & p_{k-1,n-1}(D)
            \end{bmatrix},
        $$
        and vector of feedback polynomials
        $$
            q(D) =
            \begin{bmatrix}
                q_0(D)     \\\\
                q_1(D)     \\\\
                \vdots     \\\\
                q_{k-1}(D)
            \end{bmatrix},
        $$
        the generator matrix is given by
        $$
            G(D) =
            \begin{bmatrix}
                p_{0,0}(D)/q_0(D)       & p_{0,1}(D)/q_0(D)       & \cdots & p_{0,n-1}(D)/q_0(D)       \\\\
                p_{1,0}(D)/q_1(D)       & p_{1,1}(D)/q_1(D)       & \cdots & p_{1,n-1}(D)/q_1(D)       \\\\
                \vdots                  & \vdots                  & \ddots & \vdots                    \\\\
                p_{k-1,0}(D)/q_{k-1}(D) & p_{k-1,1}(D)/q_{k-1}(D) & \cdots & p_{k-1,n-1}(D)/q_{k-1}(D)
            \end{bmatrix}.
        $$

        Examples:
            If matrix of feedforward polynomials is
            $$
                P(D) =
                \begin{bmatrix}
                    D^4 + D^2 + D + 1 && D^4 + D^3 + 1
                \end{bmatrix}
            $$
            and vector of feedback polynomials is
            $$
                q(D) =
                \begin{bmatrix}
                    D^4 + D^2 + D + 1
                \end{bmatrix},
            $$
            then the generator matrix is given by
            $$
                G(D) =
                \begin{bmatrix}
                    1 & \frac{D^4 + D^3 + 1}{D^4 + D^2 + D + 1}
                \end{bmatrix}.
            $$

                >>> code = komm.ConvolutionalCode([[0o27, 0o31]], [0o27])
                >>> for row in code.generator_matrix:
                ...     print("[" + ", ".join(str(x) for x in row) + "]")
                [0b1/0b1, 0b11001/0b10111]
        """
        n, k = self.num_output_bits, self.num_input_bits
        G_mat = np.empty((k, n), dtype=object)
        for i, j in product(range(k), range(n)):
            p = BinaryPolynomialFraction(self.feedforward_polynomials[i][j])
            q = BinaryPolynomialFraction(self.feedback_polynomials[i])
            G_mat[i, j] = p / q
        return G_mat

    @cached_property
    def constraint_lengths(self) -> Array1D[np.integer]:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]])
            >>> code.constraint_lengths
            array([4, 3])
        """
        nus = np.empty(self.num_input_bits, dtype=int)
        for i in range(self.num_input_bits):
            ps = self.feedforward_polynomials[i]
            q = self.feedback_polynomials[i]
            nus[i] = max(np.amax([p.degree for p in ps]), q.degree)
        return nus

    @cached_property
    def overall_constraint_length(self) -> int:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]])
            >>> code.overall_constraint_length
            7
        """
        return super().overall_constraint_length

    @cached_property
    def memory_order(self) -> int:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]])
            >>> code.memory_order
            4
        """
        return super().memory_order

    @cache
    def state_space_representation(
        self,
    ) -> tuple[
        Array2D[np.integer],
        Array2D[np.integer],
        Array2D[np.integer],
        Array2D[np.integer],
    ]:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]])
            >>> A_mat, B_mat, C_mat, D_mat = code.state_space_representation()
            >>> A_mat
            array([[0, 1, 0, 0, 0, 0, 0],
                   [0, 0, 1, 0, 0, 0, 0],
                   [0, 0, 0, 1, 0, 0, 0],
                   [0, 0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 0, 0, 1, 0],
                   [0, 0, 0, 0, 0, 0, 1],
                   [0, 0, 0, 0, 0, 0, 0]])
            >>> B_mat
            array([[1, 0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 0, 1, 0, 0]])
            >>> C_mat
            array([[0, 1, 0],
                   [0, 1, 0],
                   [1, 0, 0],
                   [1, 1, 0],
                   [0, 1, 0],
                   [0, 0, 1],
                   [0, 1, 1]])
            >>> D_mat
            array([[1, 1, 0],
                   [0, 0, 1]])

            >>> code = komm.ConvolutionalCode([[0o27, 0o31]], [0o27])
            >>> A_mat, B_mat, C_mat, D_mat = code.state_space_representation()
            >>> A_mat
            array([[1, 1, 0, 0],
                   [1, 0, 1, 0],
                   [0, 0, 0, 1],
                   [1, 0, 0, 0]])
            >>> B_mat
            array([[1, 0, 0, 0]])
            >>> C_mat
            array([[0, 1],
                   [0, 1],
                   [0, 1],
                   [0, 0]])
            >>> D_mat
            array([[1, 1]])
        """
        betas = [
            np.array([p.coefficients(nu + 1) for p in ps]).T
            for nu, ps in zip(self.constraint_lengths, self.feedforward_polynomials)
        ]

        alphas = [
            q.coefficients(nu + 1)[:, np.newaxis]
            for nu, q in zip(self.constraint_lengths, self.feedback_polynomials)
        ]

        A_blocks: list[npt.NDArray[np.integer]] = []
        B_blocks: list[npt.NDArray[np.integer]] = []
        C_blocks: list[npt.NDArray[np.integer]] = []
        D_blocks: list[npt.NDArray[np.integer]] = []

        for alpha, beta, nu in zip(alphas, betas, self.constraint_lengths):
            nu = int(nu)  # FIXME: Why `np.eye` does not accept `np.integer`?
            if nu == 0:
                A_blocks.append(np.zeros((0, 0), dtype=int))
            else:
                A_blocks.append(np.hstack([alpha[1:], np.eye(nu, nu - 1, dtype=int)]))
            B_blocks.append(np.eye(1, nu, dtype=int))
            C_blocks.append(beta[1:] ^ (alpha[1:] * beta[0]))
            D_blocks.append(beta[0])

        A_mat = block_diagonal(A_blocks)
        B_mat = block_diagonal(B_blocks)
        C_mat = np.vstack(C_blocks)
        D_mat = np.vstack(D_blocks)

        return A_mat, B_mat, C_mat, D_mat

    @cache
    def finite_state_machine(self) -> MealyMachine:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0b111, 0b101]])
            >>> code.finite_state_machine()
            MealyMachine(transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
                         outputs=[[0, 3], [1, 2], [3, 0], [2, 1]])
        """
        return super().finite_state_machine()

    @cache
    def is_catastrophic(self) -> bool:
        """
        Examples:
            >>> code = komm.ConvolutionalCode([[0b111, 0b101]])
            >>> code.is_catastrophic()
            False

            >>> code = komm.ConvolutionalCode([[0b11, 0b101]])
            >>> code.is_catastrophic()
            True
        """
        return super().is_catastrophic()

    @cache
    def free_distance(self) -> int:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0o31, 0o27, 0o0], [0o0, 0o12, 0o15]])
            >>> code.free_distance()
            5
        """
        return super().free_distance()

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0b111, 0b101]])
            >>> code.encode([1, 1, 1, 1])
            array([1, 1, 0, 1, 1, 0, 1, 0])
        """
        return super().encode(input)

    def encode_with_state(
        self,
        input: npt.ArrayLike,
        initial_state: npt.ArrayLike,
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
        r"""
        Examples:
            >>> code = komm.ConvolutionalCode([[0b111, 0b101]])
            >>> code.encode_with_state([1, 1, 1, 1], [0, 0])
            (array([1, 1, 0, 1, 1, 0, 1, 0]), array([1, 1]))
            >>> code.encode_with_state([1, 1, 1, 1], [1, 1])
            (array([1, 0, 1, 0, 1, 0, 1, 0]), array([1, 1]))
        """
        return super().encode_with_state(input, initial_state)



from collections.abc import Callable
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .. import abc


class UniformQuantizer(abc.ScalarQuantizer):
    r"""
    Uniform scalar quantizer. It is a [scalar quantizer](/ref/ScalarQuantizer) in which the separation between levels is a constant $\Delta$, called the *quantization step*, and the thresholds are the mid-point between adjacent levels. More precisely, the levels are given by
    $$
        v_i = (i - (L - 1)/2 + \theta) \Delta, \qquad i \in [0 : L),
    $$
    where $\theta \in \mathbb{R}$ is an arbitrary *offset* (normalized by $\Delta$), and the finite thresholds are given by
    $$
        t_i = \frac{v_{i-1} + v_i}{2}, \qquad i \in [1 : L).
    $$
    For more details, see <cite>Say06, Sec. 9.4</cite>.

    Parameters:
        num_levels: The number of quantization levels $L$. It must be greater than $1$.

        step: The quantization step $\Delta$. It must be a positive number.

        offset: The offset $\theta$ of the quantizer. The default value is $0$.

    Examples:
        >>> quantizer = komm.UniformQuantizer(num_levels=4, step=0.5)
        >>> quantizer.levels
        array([-0.75, -0.25,  0.25,  0.75])
        >>> quantizer.thresholds
        array([-0.5,  0. ,  0.5])

        >>> quantizer = komm.UniformQuantizer(num_levels=5, step=0.5)
        >>> quantizer.levels
        array([-1. , -0.5,  0. ,  0.5,  1. ])
        >>> quantizer.thresholds
        array([-0.75, -0.25,  0.25,  0.75])

        >>> quantizer = komm.UniformQuantizer(num_levels=5, step=0.5, offset=2.0)
        >>> quantizer.levels
        array([0. , 0.5, 1. , 1.5, 2. ])
        >>> quantizer.thresholds
        array([0.25, 0.75, 1.25, 1.75])
    """

    def __init__(self, num_levels: int, step: float, offset: float = 0.0):
        if not num_levels > 1:
            raise ValueError("'num_levels' must be greater than 1")
        if not step > 0:
            raise ValueError("'step' must be positive")
        self.num_levels = num_levels
        self.step = step
        self.offset = offset

    @classmethod
    def mid_riser(cls, num_levels: int, step: float):
        r"""
        Constructs a *mid-riser* uniform quantizer. In a mid-riser quantizer, $0$ is always a threshold. It is a symmetric quantizer when $L$ is even.

        Parameters:
            num_levels: The number of quantization levels $L$. It must be greater than $1$.

            step: The quantization step $\Delta$. It must be a positive number.

        Examples:
            >>> quantizer = komm.UniformQuantizer.mid_riser(num_levels=4, step=0.5)
            >>> quantizer.levels
            array([-0.75, -0.25,  0.25,  0.75])
            >>> quantizer.thresholds
            array([-0.5,  0. ,  0.5])

            >>> quantizer = komm.UniformQuantizer.mid_riser(num_levels=5, step=0.5)
            >>> quantizer.levels
            array([-1.25, -0.75, -0.25,  0.25,  0.75])
            >>> quantizer.thresholds
            array([-1. , -0.5,  0. ,  0.5])
        """
        return cls(
            num_levels=num_levels,
            step=step,
            offset=0.0 if num_levels % 2 == 0 else -0.5,
        )

    @classmethod
    def mid_tread(cls, num_levels: int, step: float):
        r"""
        Constructs a *mid-tread* uniform quantizer. In a mid-tread quantizer, $0$ is always a level. It is a symmetric quantizer when $L$ is odd.

        Parameters:
            num_levels: The number of quantization levels $L$. It must be greater than $1$.

            step: The quantization step $\Delta$. It must be a positive number.

        Examples:
            >>> quantizer = komm.UniformQuantizer.mid_tread(num_levels=4, step=0.5)
            >>> quantizer.levels
            array([-1. , -0.5,  0. ,  0.5])
            >>> quantizer.thresholds
            array([-0.75, -0.25,  0.25])

            >>> quantizer = komm.UniformQuantizer.mid_tread(num_levels=5, step=0.5)
            >>> quantizer.levels
            array([-1. , -0.5,  0. ,  0.5,  1. ])
            >>> quantizer.thresholds
            array([-0.75, -0.25,  0.25,  0.75])
        """
        return cls(
            num_levels=num_levels,
            step=step,
            offset=0.0 if num_levels % 2 == 1 else -0.5,
        )

    @cached_property
    def levels(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=8, step=3.0)
            >>> quantizer.levels
            array([-10.5,  -7.5,  -4.5,  -1.5,   1.5,   4.5,   7.5,  10.5])

            >>> quantizer = komm.UniformQuantizer(num_levels=5, step=2.0)
            >>> quantizer.levels
            array([-4., -2.,  0.,  2.,  4.])
        """
        L, θ, Δ = self.num_levels, self.offset, self.step
        return np.array([(i - (L - 1) / 2 + θ) * Δ for i in range(L)])

    @cached_property
    def thresholds(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=8, step=3.0)
            >>> quantizer.thresholds
            array([-9., -6., -3.,  0.,  3.,  6.,  9.])

            >>> quantizer = komm.UniformQuantizer(num_levels=5, step=2.0)
            >>> quantizer.thresholds
            array([-3., -1.,  1.,  3.])
        """
        return (self.levels + self.step / 2)[:-1]

    def mean_squared_error(
        self,
        input_pdf: Callable[[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        input_range: tuple[float, float],
        points_per_interval: int = 4096,
    ) -> float:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=4, step=2.0)
            >>> gaussian_pdf = lambda x: 1/np.sqrt(2*np.pi) * np.exp(-x**2/2)
            >>> quantizer.mean_squared_error(
            ...     input_pdf=gaussian_pdf,
            ...     input_range=(-5, 5),
            ... )  # doctest: +FLOAT_CMP
            0.3363025489037716
            >>> uniform_pdf = lambda x: 1/8 * (np.abs(x) <= 4)
            >>> quantizer.mean_squared_error(
            ...     input_pdf=uniform_pdf,
            ...     input_range=(-4, 4),
            ... )  # doctest: +FLOAT_CMP
            0.3333333730891729
            >>> quantizer.step**2 / 12  # doctest: +FLOAT_CMP
            0.3333333333333333
        """
        return super().mean_squared_error(input_pdf, input_range, points_per_interval)

    def digitize(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=4, step=2.0)
            >>> quantizer.digitize([-2.4,  0.8,  3.2])
            array([0, 2, 3])
        """
        return super().digitize(input)

    def quantize(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=4, step=2.0)
            >>> quantizer.quantize([-2.4,  0.8,  3.2])
            array([-3.,  1.,  3.])
        """
        return super().quantize(input)



from collections.abc import Sequence
from functools import reduce
from typing import Generic, SupportsInt, TypeVar

import numpy as np
from typing_extensions import Self

from . import field
from .BinaryPolynomial import BinaryPolynomial, default_primitive_polynomial

F = TypeVar("F", bound="FiniteBifield")


class FiniteBifieldElement(Generic[F]):
    def __init__(self, ambient: F, value: SupportsInt) -> None:
        self.ambient = ambient
        self.value = BinaryPolynomial(value)

    def __repr__(self) -> str:
        return bin(self.value.value)

    def __str__(self) -> str:
        return bin(self.value.value)

    def __hash__(self) -> int:
        return hash(self.value)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ambient == other.ambient and self.value == other.value

    def _check_same_ambient(self, other: Self):
        if self.ambient != other.ambient:
            raise ValueError("elements must belong to the same finite field")

    def __add__(self, other: Self) -> Self:
        self._check_same_ambient(other)
        return self.__class__(self.ambient, self.value + other.value)

    def __sub__(self, other: Self) -> Self:
        self._check_same_ambient(other)
        return self.__class__(self.ambient, self.value - other.value)

    def __neg__(self) -> Self:
        return self

    def __mul__(self, other: Self) -> Self:
        self._check_same_ambient(other)
        value = (self.value * other.value) % self.ambient.modulus
        return self.__class__(self.ambient, value)

    def __rmul__(self, other: int) -> Self:
        if other % 2 == 0:
            return self.__class__(self.ambient, 0)
        else:
            return self

    def inverse(self) -> Self:
        d, s, _ = BinaryPolynomial.xgcd(self.value, self.ambient.modulus)
        if d.value == 1:
            return self.__class__(self.ambient, s)
        raise ZeroDivisionError("element does not have a multiplicative inverse")

    def __truediv__(self, other: Self) -> Self:
        self._check_same_ambient(other)
        return self * other.inverse()

    def __pow__(self, exponent: int) -> Self:
        return field.power(self, exponent)

    def logarithm(self, base: Self) -> int:
        for i in range(self.ambient.order):
            if base**i == self:
                return i
        raise ValueError("element is not a power of the base")

    def conjugates(self) -> list[Self]:
        conjugate_list: list[Self] = []
        exponent = 0
        while True:
            y = self ** (2**exponent)
            if y not in conjugate_list:
                conjugate_list.append(y)
            else:
                break
            exponent += 1
        return conjugate_list

    def minimal_polynomial(self) -> BinaryPolynomial:
        one = self.ambient.one
        monomials = [np.array([y, one], dtype=object) for y in self.conjugates()]
        coefficients: list[Self] = list(reduce(np.convolve, monomials))
        return BinaryPolynomial.from_coefficients([c.value.value for c in coefficients])


class FiniteBifield:
    r"""
    Finite field with binary characteristic. Objects of this class represent a *finite field* $\mathrm{GF}(2^k)$ (also known as *Galois field*), with *characteristic* $2$ and *degree* $k$.

    Parameters:
        degree: Degree $k$ of the finite field. Must be a positive integer.

        modulus: Modulus $p(X)$ of the field, specified either as a [binary polynomial](/ref/BinaryPolynomial) or as an integer to be converted to the former. Must be an irreducible polynomial. If not specified, the modulus is chosen from [the list of default primitive polynomials](/res/primitive-polynomials).


    Examples:
        >>> field = komm.FiniteBifield(4)
        >>> field
        FiniteBifield(4)
        >>> (field.characteristic, field.degree, field.order)
        (2, 4, 16)
        >>> field.modulus
        BinaryPolynomial(0b10011)

        >>> field = komm.FiniteBifield(4, modulus=0b11001)
        >>> field
        FiniteBifield(4, modulus=0b11001)
        >>> (field.characteristic, field.degree, field.order)
        (2, 4, 16)
        >>> field.modulus
        BinaryPolynomial(0b11001)

    <h2>Construction of elements</h2>

    To construct *elements* of the finite field, call the finite field object. For example, `field(0b1101)` will construct the element whose polynomial representation is $X^3 + X^2 + 1$.

    <h2>Algebraic structure</h2>

    The following operations are supported: addition (`+`), subtraction (`-`), multiplication (`*`), division (`/`), and exponentiation (`**`).

    Examples:
        >>> field = komm.FiniteBifield(4)
        >>> x = field(0b1011)
        >>> y = field(0b1100)
        >>> x + y
        0b111
        >>> x - y
        0b111
        >>> x * y
        0b1101
        >>> x / y
        0b10
        >>> x**2
        0b1001

    <h2>Further methods on elements</h2>

    The following methods are available on elements of the finite field:

    - `logarithm(base)`: Returns the logarithm of the element, with respect to a given base.
    - `conjugates()`: Returns the conjugates of the element.
    - `minimal_polynomial()`: Returns the minimal polynomial of the element.

    For more details, see <cite>LC04, Sec. 2.5</cite>.

    Examples:
        >>> field = komm.FiniteBifield(4)
        >>> x = field(0b1011)
        >>> base = field(0b10)
        >>> x.logarithm(base)
        7
        >>> x.conjugates()
        [0b1011, 0b1001, 0b1101, 0b1110]
        >>> x.minimal_polynomial()
        BinaryPolynomial(0b11001)
    """

    def __init__(self, degree: int, modulus: BinaryPolynomial | int | None = None):
        if degree < 1:
            raise ValueError("'degree' must be a positive integer")
        self.degree = degree
        if modulus is None:
            self.modulus = default_primitive_polynomial(degree)
        else:
            self.modulus = BinaryPolynomial(modulus)
        if self.modulus.degree != self.degree:
            raise ValueError("'modulus' must have the same degree as the field")
        if not self.modulus.is_irreducible():
            raise ValueError("'modulus' must be an irreducible polynomial")

    def __call__(self, value: int | BinaryPolynomial) -> FiniteBifieldElement[Self]:
        return FiniteBifieldElement(self, value)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.degree == other.degree and self.modulus == other.modulus

    @property
    def zero(self) -> FiniteBifieldElement[Self]:
        return self(0)

    @property
    def one(self) -> FiniteBifieldElement[Self]:
        return self(1)

    @property
    def characteristic(self) -> int:
        r"""
        The characteristic $2$ of the finite field.
        """
        return 2

    @property
    def order(self) -> int:
        r"""
        The order (number of elements) of the finite field. It is given by $2^k$.
        """
        return 2**self.degree

    def __repr__(self) -> str:
        if self.modulus.value == default_primitive_polynomial(self.degree):
            args = f"{self.degree}"
        else:
            args = f"{self.degree}, modulus={self.modulus}"
        return f"{self.__class__.__name__}({args})"


def find_roots(
    field: F,
    coefficients: Sequence[FiniteBifieldElement[F]],
) -> list[FiniteBifieldElement[F]]:
    r"""
    Returns the roots of a polynomial with coefficients in a finite field. This function uses exhaustive search to find the roots.

    Parameters:
        field: Finite field.
        coefficients: Coefficients of the polynomial, in increasing order of degree.

    Returns:
        List of roots of the polynomial.

    Examples:
        >>> field = komm.FiniteBifield(4)
        >>> alpha = field(0b10)  # alpha = X, a primitive element
        >>> coefficients = [field.one, field.one, field.zero, alpha**5]  # 1 + X + alpha^5 X^3
        >>> find_roots(field, coefficients)
        [0b111, 0b1000, 0b1111]
        >>> [alpha**10, alpha**3, alpha**12]
        [0b111, 0b1000, 0b1111]
    """
    roots: list[FiniteBifieldElement[F]] = []
    for i in range(field.order):
        x = field(i)
        evaluated = field.zero
        for coefficient in reversed(coefficients):  # Horner's method
            evaluated = evaluated * x + coefficient
        if evaluated == field.zero:
            roots.append(x)
            if len(roots) >= len(coefficients) - 1:
                break
    return roots



from functools import cache

import numpy as np
import numpy.typing as npt

from .. import abc
from .ProductLabeling import ProductLabeling
from .ReflectedLabeling import ReflectedLabeling


class ReflectedRectangularLabeling(abc.Labeling):
    r"""
    Reflected rectangular binary labeling. It is the [Cartesian product](/ref/ProductLabeling) of two [reflected binary labelings](/ref/ReflectedLabeling), possibly with distinct number of bits.
    """

    def __init__(
        self,
        num_bits: int | tuple[int, int],
        _pre_cache: bool = True,
    ) -> None:
        if isinstance(num_bits, int):
            if not (num_bits > 0 and num_bits % 2 == 0):
                raise ValueError(
                    "if a single integer, 'num_bits' must be an even number"
                )
            num_bits = (num_bits // 2, num_bits // 2)
        if not (num_bits[0] >= 1 and num_bits[1] >= 1):
            raise ValueError("'num_bits' must contain positive integers")
        self._num_bits = num_bits
        if _pre_cache:
            self.matrix

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._num_bits})"

    @property
    @cache
    def matrix(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedRectangularLabeling(4)
            >>> labeling.matrix
            array([[0, 0, 0, 0],
                   [0, 0, 0, 1],
                   [0, 0, 1, 1],
                   [0, 0, 1, 0],
                   [0, 1, 0, 0],
                   [0, 1, 0, 1],
                   [0, 1, 1, 1],
                   [0, 1, 1, 0],
                   [1, 1, 0, 0],
                   [1, 1, 0, 1],
                   [1, 1, 1, 1],
                   [1, 1, 1, 0],
                   [1, 0, 0, 0],
                   [1, 0, 0, 1],
                   [1, 0, 1, 1],
                   [1, 0, 1, 0]])
        """
        mi, mq = self._num_bits
        matrix = ProductLabeling(ReflectedLabeling(mi), ReflectedLabeling(mq)).matrix
        return matrix

    @property
    def num_bits(self) -> int:
        r"""
        Examples:
            >>> labeling = komm.ReflectedRectangularLabeling(4)
            >>> labeling.num_bits
            4
        """
        mi, mq = self._num_bits
        return mi + mq

    @property
    def cardinality(self) -> int:
        r"""
        Examples:
            >>> labeling = komm.ReflectedRectangularLabeling(4)
            >>> labeling.cardinality
            16
        """
        mi, mq = self._num_bits
        return 2 ** (mi + mq)

    @property
    # @cache
    def inverse_mapping(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedRectangularLabeling(4  )
            >>> labeling.inverse_mapping
            {(0, 0, 0, 0): 0,
             (0, 0, 0, 1): 1,
             (0, 0, 1, 1): 2,
             (0, 0, 1, 0): 3,
             (0, 1, 0, 0): 4,
             (0, 1, 0, 1): 5,
             (0, 1, 1, 1): 6,
             (0, 1, 1, 0): 7,
             (1, 1, 0, 0): 8,
             (1, 1, 0, 1): 9,
             (1, 1, 1, 1): 10,
             (1, 1, 1, 0): 11,
             (1, 0, 0, 0): 12,
             (1, 0, 0, 1): 13,
             (1, 0, 1, 1): 14,
             (1, 0, 1, 0): 15}
        """
        return super().inverse_mapping

    def indices_to_bits(self, indices: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedRectangularLabeling(4)
            >>> labeling.indices_to_bits([8, 13])
            array([1, 1, 0, 0, 1, 0, 0, 1])
            >>> labeling.indices_to_bits([[8, 13], [0, 1]])
            array([[1, 1, 0, 0, 1, 0, 0, 1],
                   [0, 0, 0, 0, 0, 0, 0, 1]])
        """
        return super().indices_to_bits(indices)

    def bits_to_indices(self, bits: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedRectangularLabeling(4)
            >>> labeling.bits_to_indices([1, 1, 0, 0, 1, 0, 0, 1])
            array([ 8, 13])
            >>> labeling.bits_to_indices([
            ...     [1, 1, 0, 0, 1, 0, 0, 1],
            ...     [0, 0, 0, 0, 0, 0, 0, 1],
            ... ])
            array([[ 8, 13],
                   [ 0,  1]])
        """
        return super().bits_to_indices(bits)

    def marginalize(self, metrics: npt.ArrayLike) -> npt.NDArray[np.floating]:
        return super().marginalize(metrics)



from typing import Any

import numpy as np
import numpy.typing as npt


def sampling_rate_compress(
    input: npt.ArrayLike,
    factor: int,
    offset: int = 0,
    axis: int = -1,
) -> npt.NDArray[Any]:
    r"""
    Performs sampling rate compression (downsampling). For a given input $x[n]$, the output is
    $$
    y[n] = x[n M + \Delta],
    $$
    where $M$ is the *compression factor* and $\Delta \in [0:M)$ is the *offset* for the first selected element. In words, the compressor extracts every $M$-th element starting from the offset $\Delta$ along the specified axis. For more details, see <cite>OS99, Sec. 4.6.1</cite>.

    Parameters:
        input: The input array $x[n]$ to be compressed.
        factor: The compression factor $M$.
        offset: The offset $\Delta$. Must satisfy $\Delta \in [0:M)$.
        axis: The axis along which to extract elements. Default is the last axis.

    Returns:
        output: The compressed array $y[n]$.

    Examples:
        >>> komm.sampling_rate_compress(
        ...    [[11, 12, 13, 14, 15],
        ...     [16, 17, 18, 19, 20]],
        ...    factor=3,
        ... )
        array([[11, 14],
               [16, 19]])
        >>> komm.sampling_rate_compress(
        ...    [[11, 12, 13, 14, 15],
        ...     [16, 17, 18, 19, 20]],
        ...    factor=3,
        ...    offset=1,
        ... )
        array([[12, 15],
               [17, 20]])
        >>> komm.sampling_rate_compress(
        ...     [[11, 12],
        ...      [13, 14],
        ...      [15, 16],
        ...      [17, 18],
        ...      [19, 20]],
        ...     factor=3,
        ...     axis=0,
        ... )
        array([[11, 12],
               [17, 18]])
    """
    if not factor > 0:
        raise ValueError("'factor' should be a positive integer")
    if not (0 <= offset < factor):
        raise ValueError("'offset' should satisfy 0 <= offset < factor")
    input = np.asarray(input)
    indexer = [slice(None)] * input.ndim
    indexer[axis] = slice(offset, input.shape[axis], factor)
    return input[tuple(indexer)]


def sampling_rate_expand(
    input: npt.ArrayLike,
    factor: int,
    offset: int = 0,
    axis: int = -1,
) -> npt.NDArray[Any]:
    r"""
    Performs sampling rate expansion (upsampling). For a given input $x[n]$, the output is
    $$
    y[n] = \begin{cases}
    x[n \operatorname{div} L] & \text{if } n \bmod L = \Delta, \\\\
    0,       & \text{otherwise}
    \end{cases}
    $$
    where $L$ is the *expansion factor* and $\Delta \in [0:L)$ is the *offset* for the first output element. In words, the expander inserts $L-1$ zeros between each element of the input array along the specified axis, starting from the offset $\Delta$. For more details, see <cite>OS99, Sec. 4.6.2</cite>.

    Parameters:
        input: The input array $x[n]$ to be expanded.
        factor: The expansion factor $L$.
        offset: The offset $\Delta$. Must satisfy $\Delta \in [0:L)$.
        axis: The axis along which to insert zeros. Default is the last axis.

    Returns:
        output: The expanded array $y[n]$.

    Examples:
        >>> komm.sampling_rate_expand([[1, 2], [3, 4]], factor=3)
        array([[1, 0, 0, 2, 0, 0],
               [3, 0, 0, 4, 0, 0]])
        >>> komm.sampling_rate_expand([[1, 2], [3, 4]], factor=3, offset=1)
        array([[0, 1, 0, 0, 2, 0],
               [0, 3, 0, 0, 4, 0]])
        >>> komm.sampling_rate_expand([[1, 2], [3, 4]], factor=3, axis=0)
        array([[1, 2],
               [0, 0],
               [0, 0],
               [3, 4],
               [0, 0],
               [0, 0]])
    """
    if not factor > 0:
        raise ValueError("'factor' should be a positive integer")
    if not (0 <= offset < factor):
        raise ValueError("'offset' should satisfy 0 <= offset < factor")
    input = np.asarray(input)
    shape = list(input.shape)
    shape[axis] *= factor
    output = np.zeros(shape, dtype=input.dtype)
    indexer = [slice(None)] * input.ndim
    indexer[axis] = slice(offset, shape[axis], factor)
    output[tuple(indexer)] = input
    return output


def fourier_transform(
    waveform: npt.ArrayLike,
    time_step: float,
    nfft: int | None = None,
    axis: int = -1,
) -> tuple[npt.NDArray[np.complexfloating], npt.NDArray[np.floating]]:
    r"""
    Computes the Fourier transform. This function applies a shift to the spectrum (so that the zero frequency component is at the center) and scales the output by a given time step. Both the spectrum and the corresponding frequency bins are returned.

    Note:
        This is a simple wrapper around `numpy.fft` functions.

    Parameters:
        waveform: The input array representing the waveform to be transformed.
        time_step: The time step between samples in the waveform.
        nfft: The number of points in the FFT. If `None`, it defaults to the size of the input along the specified axis.
        axis: The axis along which to compute the Fourier transform. Default is the last axis.

    Returns:
        spectrum: The spectrum correponding to the input waveform.
        frequencies: The corresponding frequency bins.

    Examples:
        >>> spectrum, frequencies = komm.fourier_transform([1, 2, 3, 4], time_step=0.1)
        >>> spectrum
        array([-0.2+0.j , -0.2-0.2j,  1. +0.j , -0.2+0.2j])
        >>> frequencies
        array([-5. , -2.5,  0. ,  2.5])
    """
    waveform = np.asarray(waveform)
    n = nfft or waveform.shape[axis]
    spectrum = np.fft.fftshift(np.fft.fft(waveform, n=n, axis=axis)) * time_step
    frequencies = np.fft.fftshift(np.fft.fftfreq(n=n, d=time_step))
    return spectrum, frequencies



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

import numpy as np
import numpy.typing as npt

from .._util.validators import validate_pmf
from ..types import Array1D, Array2D

T = TypeVar("T", np.floating, np.complexfloating)


class Constellation(ABC, Generic[T]):
    @property
    @abstractmethod
    def matrix(self) -> Array2D[T]:
        r"""
        The constellation matrix $\mathbf{X}$.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def order(self) -> int:
        r"""
        The order $M$ of the constellation.
        """
        return self.matrix.shape[0]

    @property
    @abstractmethod
    def dimension(self) -> int:
        r"""
        The dimension $N$ of the constellation.
        """
        return self.matrix.shape[1]

    @abstractmethod
    def mean(self, priors: npt.ArrayLike | None = None) -> Array1D[T]:
        r"""
        Computes the mean $\mathbf{m}$ of the constellation given prior probabilities $p_i$ of the constellation symbols. It is given by
        $$
        \mathbf{m} = \sum_{i \in [0:M)} p_i \mathbf{x}_i.
        $$

        Parameters:
            priors: The prior probabilities of the constellation symbols. Must be a 1D-array whose size is equal to the order $M$ of the constellation. If not given, uniform priors are assumed.

        Returns:
            mean: The mean $\mathbf{m}$ of the constellation.
        """
        if priors is None:
            priors = np.ones(self.order) / self.order
        priors = validate_pmf(priors)
        return np.dot(priors, self.matrix)

    @abstractmethod
    def mean_energy(self, priors: npt.ArrayLike | None = None) -> np.floating:
        r"""
        Computes the mean energy $E$ of the constellation given prior probabilities $p_i$ of the constellation symbols. It is given by
        $$
        E = \sum_{i \in [0:M)} p_i \lVert \mathbf{x}_i \rVert^2.
        $$

        Parameters:
            priors: The prior probabilities of the constellation symbols. Must be a 1D-array whose size is equal to the order $M$ of the constellation. If not given, uniform priors are assumed.

        Returns:
            mean_energy: The mean energy $E$ of the constellation.
        """
        if priors is None:
            priors = np.ones(self.order) / self.order
        priors = validate_pmf(priors)
        return priors @ np.sum(np.abs(self.matrix) ** 2, axis=1)

    @abstractmethod
    def minimum_distance(self) -> np.floating:
        r"""
        Computes the minimum Euclidean distance $d_\mathrm{min}$ of the constellation. It is given by
        $$
        d_\mathrm{min} = \min_ { i, j \in [0:M), ~ i \neq j } \lVert \mathrm{x}_i - \mathrm{x}_j \rVert.
        $$
        """
        distances = np.linalg.norm(self.matrix[:, np.newaxis] - self.matrix, axis=2)
        np.fill_diagonal(distances, np.inf)  # Ignore self-distances
        return np.min(distances)

    @abstractmethod
    def indices_to_symbols(self, indices: npt.ArrayLike) -> npt.NDArray[T]:
        r"""
        Returns the constellation symbols corresponding to the given indices.

        Parameters:
            indices: The indices to be converted to symbols. Must be an array of integers in $[0:M)$.

        Returns:
            symbols: The symbols corresponding to the given indices. Has the same shape as `indices`, but with the last dimension expanded by a factor of $N$.
        """
        indices = np.asarray(indices, dtype=int)
        symbols = self.matrix[indices].reshape(*indices.shape[:-1], -1)
        return symbols

    def _squared_distances(self, input: npt.NDArray[T]) -> npt.NDArray[np.floating]:
        diff = input[..., np.newaxis, :] - self.matrix[np.newaxis, ...]
        return np.sum(np.abs(diff) ** 2, axis=-1)

    @abstractmethod
    def closest_indices(self, received: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Returns the indices of the constellation symbols closest to the given received points.

        Parameters:
            received: The received points. Must be an array whose last dimension is a multiple of $N$.

        Returns:
            indices: The indices of the symbols closest to the received points. Has the same shape as `received`, but with the last dimension contracted by a factor of $N$.
        """
        N = self.dimension
        received = np.asarray(received)
        if received.shape[-1] % N != 0:
            raise ValueError(
                "last dimension of 'received' must be a multiple of the constellation"
                f" dimension {N} (got {received.shape[-1]})"
            )
        distances = self._squared_distances(received.reshape(-1, N))
        indices = np.argmin(distances, axis=-1).reshape(*received.shape[:-1], -1)
        return indices

    @abstractmethod
    def closest_symbols(self, received: npt.ArrayLike) -> npt.NDArray[T]:
        r"""
        Returns the constellation symbols closest to the given received points.

        Parameters:
            received: The received points. Must be an array whose last dimension is a multiple of $N$.

        Returns:
            symbols: The symbols closest to the received points. Has the same shape as `received`.
        """
        received = np.asarray(received)
        indices = self.closest_indices(received)
        symbols = self.indices_to_symbols(indices).reshape(*received.shape[:-1], -1)
        return symbols

    @abstractmethod
    def posteriors(
        self,
        received: npt.ArrayLike,
        snr: float,
        priors: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Returns the posterior probabilities of each constellation symbol given received points, the signal-to-noise ratio (SNR), and prior probabilities.

        Parameters:
            received: The received points. Must be an array whose last dimension is a multiple of $N$.

            snr: The signal-to-noise ratio (SNR) of the channel (linear, not decibel).

            priors: The prior probabilities of the symbols. Must be a 1D-array whose size is equal to $M$. If not given, uniform priors are assumed.

        Returns:
            posteriors: The posterior probabilities of each symbol given the received points. Has the same shape as `received`, but with the last dimension changed by a factor of $M / N$.
        """
        M, N = self.order, self.dimension
        if priors is None:
            priors = np.ones(M) / M
        priors = validate_pmf(priors)
        if priors.size != M:
            raise ValueError(
                "length of 'priors' must be equal to the constellation"
                f" order {M} (got {priors.size})"
            )
        received = np.asarray(received)
        if received.shape[-1] % N != 0:
            raise ValueError(
                "last dimension of 'received' must be a multiple of the constellation"
                f" dimension {N} (got {received.shape[-1]})"
            )
        r = received.reshape(-1, N)
        n0 = self.mean_energy(priors) / snr
        distances = self._squared_distances(r)
        logp = -distances / n0 + np.log(priors)
        logp -= np.max(logp, axis=-1, keepdims=True)
        p = np.exp(logp)
        p /= np.sum(p, axis=-1, keepdims=True)
        return p.reshape(*received.shape[:-1], -1)



from collections.abc import Sequence
from itertools import product
from typing import Any

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from komm._algebra.BinaryPolynomial import BinaryPolynomial

ArrayInt = npt.NDArray[np.integer]


def _pivot(row: ArrayInt) -> int:
    f_list = np.flatnonzero(row)
    return f_list[0] if f_list.size > 0 else row.size


def rref(matrix: npt.ArrayLike) -> ArrayInt:
    r"""
    Computes the row-reduced echelon form of a matrix in $\ZZ_2$.

    Parameters:
        matrix: The matrix to reduce. Its elements must be `0` or `1`.

    Returns:
        reduced: The row-reduced echelon form of the matrix.

    Examples:
        >>> rref([[1, 1, 1], [1, 1, 0], [0, 0, 1]])
        array([[1, 1, 0],
               [0, 0, 1],
               [0, 0, 0]])

    References:
        [1] https://gist.github.com/rgov/1499136
    """
    reduced = np.asarray(matrix, dtype=int)
    n_rows, n_cols = reduced.shape
    for r in tqdm(range(n_rows), desc="Row-reducing matrix", unit="row", delay=2.5):
        # Choose the pivot
        possible_pivots = [_pivot(row) for row in reduced[r:]]
        p = np.argmin(possible_pivots) + r
        # Swap rows
        reduced[r], reduced[p] = reduced[p].copy(), reduced[r].copy()
        # Pivot column
        f = _pivot(reduced[r])
        if f >= n_cols:
            continue
        # Subtract the row from others
        for i in range(n_rows):
            if i != r and reduced[i, f] != 0:
                reduced[i] = (reduced[i] + reduced[r]) % 2
    return reduced


def xrref(matrix: npt.ArrayLike) -> tuple[ArrayInt, ArrayInt, ArrayInt]:
    r"""
    Computes the row-reduced echelon form of a matrix in $\ZZ_2$. Returns the row transformation matrix, the reduced matrix, and the pivot indices. The relation between the input matrix `matrix` and the reduced matrix `reduced` is given by the equation `reduced = row_transform @ matrix`.

    Parameters:
        matrix: The matrix to reduce. Its elements must be `0` or `1`.

    Returns:
        row_transform: The row transformation matrix.
        reduced: The row-reduced echelon form of the matrix.
        pivots: The pivot indices.

    Examples:
        >>> matrix = np.array([[1, 1, 1], [1, 1, 0], [0, 0, 1]])
        >>> row_transform, reduced, pivots = xrref(matrix)
        >>> row_transform
        array([[0, 1, 0],
               [0, 0, 1],
               [1, 1, 1]])
        >>> reduced
        array([[1, 1, 0],
               [0, 0, 1],
               [0, 0, 0]])
        >>> pivots
        array([0, 2])
        >>> np.array_equal((row_transform @ matrix) % 2, reduced)
        True
    """
    matrix = np.asarray(matrix, dtype=int)
    n_rows, n_cols = matrix.shape
    eye = np.eye(n_rows, dtype=int)
    augmented = np.concatenate((matrix, eye), axis=1)
    augmented_rref = rref(augmented)

    reduced = augmented_rref[:, :n_cols]
    row_transform = augmented_rref[:, n_cols:]

    pivots: list[int] = []
    for j, col in enumerate(reduced.T):
        if len(pivots) >= n_rows:
            break
        if np.array_equal(col, eye[len(pivots)]):
            pivots.append(j)

    return row_transform, reduced, np.array(pivots)


def rank(matrix: npt.ArrayLike) -> int:
    r"""
    Computes the rank of a matrix in $\ZZ_2$.

    Parameters:
        matrix: The matrix. Its elements must be `0` or `1`.

    Returns:
        rank: The rank of the matrix.

    Examples:
        >>> rank([[1, 1, 1], [1, 1, 0], [0, 0, 1]])
        2
    """
    _, _, pivots = xrref(matrix)
    return len(pivots)


def pseudo_inverse(matrix: npt.ArrayLike) -> ArrayInt:
    r"""
    Computes a pseudo inverse of a matrix in $\ZZ_2$.

    Parameters:
        matrix: The matrix. Its elements must be `0` or `1`.

    Returns:
        p_inverse: A pseudo inverse of the matrix.

    Examples:
        >>> matrix = np.array([[1, 0, 1], [1, 1, 1]])
        >>> p_inverse = pseudo_inverse(matrix)
        >>> p_inverse
        array([[1, 0],
               [1, 1],
               [0, 0]])
        >>> (matrix @ p_inverse) % 2
        array([[1, 0],
               [0, 1]])

        >>> matrix = np.array([[1, 1, 1], [1, 1, 0], [0, 0, 1]])
        >>> p_inverse = pseudo_inverse(matrix)
        >>> p_inverse
        array([[0, 1, 0],
               [0, 0, 0],
               [0, 0, 1]])
        >>> np.array_equal((matrix @ p_inverse @ matrix) % 2, matrix)
        True
        >>> np.array_equal((p_inverse @ matrix @ p_inverse) % 2, p_inverse)
        True
    """
    row_transform, reduced, pivots = xrref(matrix)
    reduced_inverse = np.zeros_like(reduced.T)
    if pivots.size == 0:
        return reduced_inverse
    reduced_inverse[pivots] = np.eye(pivots.size, reduced.shape[0])
    p_inverse = np.dot(reduced_inverse, row_transform) % 2
    return p_inverse


def null_matrix(matrix: npt.ArrayLike) -> ArrayInt:
    r"""
    Computes a null matrix of a matrix in $\ZZ_2$.
    """
    _, reduced, pivots = xrref(matrix)
    n_rows, n_cols = reduced.shape
    null = np.empty((n_cols - n_rows, n_cols), dtype=int)
    not_pivots = np.setdiff1d(range(n_cols), pivots)
    null[:, not_pivots] = np.eye(n_cols - n_rows, dtype=int)
    null[:, pivots] = reduced[:, not_pivots].T
    return null


def block_diagonal(arrays: Sequence[npt.ArrayLike]) -> npt.NDArray[Any]:
    arrays_np = [np.asarray(a) for a in arrays]
    total_rows = sum(a.shape[0] for a in arrays_np)
    total_cols = sum(a.shape[1] for a in arrays_np)
    result = np.zeros((total_rows, total_cols), dtype=arrays_np[0].dtype)
    r_off, c_off = 0, 0
    for a in arrays_np:
        r, c = a.shape
        result[r_off : r_off + r, c_off : c_off + c] = a
        r_off += r
        c_off += c
    return result


def invariant_factors(matrix: npt.ArrayLike) -> list[BinaryPolynomial]:
    r"""
    Computes the invariant factors of a matrix over $\ZZ_2[D]$.

    Parameters:
        matrix: The matrix. Its elements must be [binary polynomials](/ref/BinaryPolynomial) or integers to be converted to the former.

    Returns:
        factors: The invariant factors of the matrix.

    Examples:
        >>> invariant_factors([[0b1, 0b111, 0b101, 0b11], [0b10, 0b111, 0b100, 0b1]])
        [BinaryPolynomial(0b1), BinaryPolynomial(0b111)]
    """
    # See [McE98, Appendix B, p. 1128]
    BP = BinaryPolynomial
    matrix = np.array(matrix, dtype=int)
    k, n = matrix.shape
    if k > n:
        raise ValueError("number of rows cannot exceed number of columns")
    if np.all(matrix == 0):
        return []

    # E1: Find entry of least size and move it to position (0,0)
    i0, j0, d0, a0 = -1, -1, np.inf, BP(0b0)
    for i, j in product(range(k), range(n)):
        a = BP(matrix[i, j])
        if a == 0b0:
            continue
        d = a.degree
        if d < d0:
            i0, j0, d0, a0 = i, j, d, a
    matrix[0, :], matrix[i0, :] = matrix[i0, :].copy(), matrix[0, :].copy()
    matrix[:, 0], matrix[:, j0] = matrix[:, j0].copy(), matrix[:, 0].copy()

    # E2a
    done = False
    while not done:
        done = True
        for j in range(1, n):
            a = BP(matrix[0, j])
            if a == 0:
                continue
            q, r = divmod(a, a0)
            if r != 0:
                a0 = r
                for i in range(k):
                    matrix[i, j] = BP(matrix[i, j]) - q * BP(matrix[i, 0])
                matrix[:, 0], matrix[:, j] = matrix[:, j].copy(), matrix[:, 0].copy()
                done = False
                break
    # E2b
    done = False
    while not done:
        done = True
        for i in range(1, k):
            a = BP(matrix[i, 0])
            if a == 0:
                continue
            q, r = divmod(a, a0)
            if r != 0:
                a0 = r
                for j in range(n):
                    matrix[i, j] = BP(matrix[i, j]) - q * BP(matrix[0, j])
                matrix[0, :], matrix[i, :] = matrix[i, :].copy(), matrix[0, :].copy()
                done = False
                break

    # E3a
    for j in range(1, n):
        q = BP(matrix[0, j]) // a0
        for i in range(k):
            matrix[i, j] = BP(matrix[i, j]) - q * BP(matrix[i, 0])
    # E3b
    for i in range(1, k):
        q = BP(matrix[i, 0]) // a0
        for j in range(n):
            matrix[i, j] = BP(matrix[i, j]) - q * BP(matrix[0, j])

    return [a0] + invariant_factors(matrix[1:, 1:])



from functools import cache, cached_property

import numpy as np
import numpy.typing as npt

from .. import abc
from .._algebra.BinaryPolynomial import BinaryPolynomial
from .._algebra.BinaryPolynomialFraction import BinaryPolynomialFraction
from .._finite_state_machine.MealyMachine import MealyMachine
from .._util.format import format_list_no_quotes as fmt
from ..types import Array1D, Array2D


class HighRateConvolutionalCode(abc.ConvolutionalCode):
    r"""
    High-rate convolutional encoder. It is an $(n, n-1)$ recursive systematic [convolutional encoder](/ref/ConvolutionalCode) defined by a single *check row* $h(D) \in \mathbb{F}_2[D]^n$ and realized in observable canonical form. By convention, the first $n - 1$ positions represent the information bits.

    Parameters:
        h_row: The check row $h(D)$ of the encoder. Must be an $n$-vector whose entries are either [binary polynomials](/ref/BinaryPolynomial) or integers to be converted to the former.

    Examples:
        Consider the high-rate convolutional encoder with $(n, k, \sigma) = (4, 3, 3)$ depicted below.

        <figure markdown>
        ![Convolutional encoder for high-rate (4, 3, 3) code.](/fig/cc_high_rate_4_3_3.svg)
        </figure>

        Its check row is given by
        $$
            h(D) =
            \begin{bmatrix}
                D^3 + D  &&  D^3 + D^2 + 1  &&  D^3 + D + 1  &&  D^3 + 1
            \end{bmatrix}.
        $$

            >>> code = komm.HighRateConvolutionalCode([0b1010, 0b1101, 0b1011, 0b1001])
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])

    Please refer to [the table of optimal high-rate convolutional codes](/res/convolutional-codes/#high-rate).
    """

    h_row: list[BinaryPolynomial]

    def __init__(self, h_row: npt.ArrayLike) -> None:
        h_row = np.asarray(h_row, dtype=int)
        if h_row.ndim != 1:
            raise ValueError("'h_row' must be a 1-dimensional array")
        self.h_row = [BinaryPolynomial(x) for x in h_row]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({fmt(self.h_row)})"

    @cached_property
    def num_input_bits(self) -> int:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.num_input_bits
            3
        """
        return len(self.h_row) - 1

    @cached_property
    def num_output_bits(self) -> int:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.num_output_bits
            4
        """
        return len(self.h_row)

    @cached_property
    def degree(self) -> int:
        r"""
        For a high-rate convolutional encoder realized in observable canonical form, the degree $\sigma$ is the maximum degree of the polynomials in the check row $h(D)$.

        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.degree
            3
        """
        return max(p.degree for p in self.h_row)

    @cached_property
    def generator_matrix(self) -> Array2D[np.object_]:
        r"""
        For a high-rate convolutional code with check row
        $$
            h(D) =
            \begin{bmatrix}
                h_0(D)  &&  h_1(D)  &&  \cdots  &&  h_{n-1}(D)
            \end{bmatrix},
        $$
        the generator matrix is given by
        $$
            G(D) =
            \begin{bmatrix}
                1      & 0      & \cdots & 0      & h_0(D) / h_{n-1}(D)     \\\\[1ex]
                0      & 1      & \cdots & 0      & h_1(D) / h_{n-1}(D)     \\\\[1ex]
                \vdots & \vdots & \ddots & \vdots & \vdots                  \\\\[1ex]
                0      & 0      & \cdots & 1      & h_{n-2}(D) / h_{n-1}(D)
            \end{bmatrix}.
        $$

        Examples:
            If the check row is
            $$
                h(D) =
                \begin{bmatrix}
                    D^3 + D  &&  D^3 + D^2 + 1  &&  D^3 + D + 1  &&  D^3 + 1
                \end{bmatrix},
            $$
            then the generator matrix is
            $$
                G(D) =
                \begin{bmatrix}
                    1 & 0 & 0 & \frac{D^3 + D}{D^3 + 1} \\\\[1ex]
                    0 & 1 & 0 & \frac{D^3 + D^2 + 1}{D^3 + 1} \\\\[1ex]
                    0 & 0 & 1 & \frac{D^3 + D + 1}{D^3 + 1}
                \end{bmatrix}.
            $$

                >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
                >>> for row in code.generator_matrix:
                ...     print("[" + ", ".join(str(x).ljust(12) for x in row) + "]")
                [0b1/0b1     , 0b0/0b1     , 0b0/0b1     , 0b110/0b111 ]
                [0b0/0b1     , 0b1/0b1     , 0b0/0b1     , 0b1101/0b1001]
                [0b0/0b1     , 0b0/0b1     , 0b1/0b1     , 0b1011/0b1001]
        """
        n, k = self.num_output_bits, self.num_input_bits
        G_mat = np.empty((k, n), dtype=object)
        for i in range(n - 1):
            for j in range(n - 1):
                G_mat[i, j] = BinaryPolynomialFraction(int(i == j))
            G_mat[i, -1] = BinaryPolynomialFraction(self.h_row[i], self.h_row[-1])
        return G_mat

    @cached_property
    def constraint_lengths(self) -> Array1D[np.integer]:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.constraint_lengths
            array([3, 3, 3])
        """
        ps = self.h_row[:-1]
        q = self.h_row[-1]
        return np.array([max(p.degree, q.degree) for p in ps], dtype=int)

    @cached_property
    def overall_constraint_length(self) -> int:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.overall_constraint_length
            9
        """
        return super().overall_constraint_length

    @cached_property
    def memory_order(self) -> int:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.memory_order
            3
        """
        return super().memory_order

    @cache
    def state_space_representation(
        self,
    ) -> tuple[
        Array2D[np.integer],
        Array2D[np.integer],
        Array2D[np.integer],
        Array2D[np.integer],
    ]:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> A_mat, B_mat, C_mat, D_mat = code.state_space_representation()
            >>> A_mat
            array([[0, 1, 0],
                   [0, 0, 1],
                   [1, 0, 0]])
            >>> B_mat
            array([[1, 0, 1],
                   [0, 1, 0],
                   [0, 0, 1]])
            >>> C_mat
            array([[0, 0, 0, 0],
                   [0, 0, 0, 0],
                   [0, 0, 0, 1]])
            >>> D_mat
            array([[1, 0, 0, 0],
                   [0, 1, 0, 1],
                   [0, 0, 1, 1]])
        """
        n, σ = self.num_output_bits, self.degree
        beta = np.array([p.coefficients(σ + 1) for p in self.h_row[:-1]])
        beta = np.fliplr(beta)
        alpha = np.array([self.h_row[-1].coefficients(σ + 1)])
        alpha = np.fliplr(alpha)
        A_mat = np.vstack([np.eye(σ - 1, σ, k=1, dtype=int), alpha[:, :-1]])
        B_mat = np.empty((n - 1, σ), dtype=int)
        for j in range(σ):
            B_mat[:, j] = beta[:, j] ^ (alpha[0, j] * beta[:, -1])
        C_mat = np.zeros((σ, n), dtype=int)
        C_mat[-1, -1] = 1
        D_mat = np.hstack([np.eye(n - 1, dtype=int), beta[:, -1:]])
        return A_mat, B_mat, C_mat, D_mat

    @cache
    def finite_state_machine(self) -> MealyMachine:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> fsm = code.finite_state_machine()
            >>> (fsm.num_input_symbols, fsm.num_output_symbols, fsm.num_states)
            (8, 16, 8)
        """
        return super().finite_state_machine()

    @cache
    def is_catastrophic(self) -> bool:
        """
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.is_catastrophic()
            False
        """
        return super().is_catastrophic()

    @cache
    def free_distance(self) -> int:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.free_distance()
            4
        """
        return super().free_distance()

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.encode([1, 1, 1, 1, 1, 1])
            array([1, 1, 1, 0, 1, 1, 1, 0])
        """
        return super().encode(input)

    def encode_with_state(
        self,
        input: npt.ArrayLike,
        initial_state: npt.ArrayLike,
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
        r"""
        Examples:
            >>> code = komm.HighRateConvolutionalCode([0o12, 0o15, 0o13, 0o11])
            >>> code.encode_with_state([1, 1, 1, 1, 1, 1], [0, 0, 0])
            (array([1, 1, 1, 0, 1, 1, 1, 0]), array([1, 0, 1]))
            >>> code.encode_with_state([1, 1, 1, 1, 1, 1], [1, 0, 1])
            (array([1, 1, 1, 1, 1, 1, 1, 0]), array([1, 1, 0]))
        """
        return super().encode_with_state(input, initial_state)



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .. import abc
from .util import rect, tri


@dataclass
class RectangularPulse(abc.Pulse):
    r"""
    Rectangular pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        p(t) =
        \begin{cases}
            1, & 0 \leq t < w, \\\\
            0, & \text{otherwise},
        \end{cases}
    $$
    where $w$ is the *relative width* of the pulse, which must satisfy $0 < w \leq 1$.

    The waveform of the rectangular pulse is depicted below for $w = 1$, and for $w = 0.5$.

    <div class="centered" markdown>
      <span>
      ![Rectangular NRZ pulse.](/fig/pulse_rectangular_nrz.svg)
      </span>
      <span>
      ![Rectangular RZ pulse.](/fig/pulse_rectangular_rz.svg)
      </span>
    </div>

    Notes:
        - For $w = 1$ it is also called the _NRZ pulse_.
        - For $w = 0.5$ it is also called the _halfway RZ pulse_.

    Parameters:
        width: The relative width $w$ of the pulse. Must satisfy $0 < w \leq 1$. The default value is `1.0`.
    """

    width: float = 1.0

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the rectangular pulse, it is given by
        $$
            p(t) = \rect\left(\frac{t - w/2}{w}\right).
        $$

        Examples:
            >>> pulse = komm.RectangularPulse(width=1.0)  # NRZ pulse
            >>> pulse.waveform(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )
            array([0., 0., 0., 0., 1., 1., 1., 1., 0.])

            >>> pulse = komm.RectangularPulse(width=0.5)  # Halfway RZ pulse
            >>> pulse.waveform(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )
            array([0., 0., 0., 0., 1., 1., 0., 0., 0.])
        """
        w = self.width
        t = np.asarray(t)
        return rect((t - w / 2) / w)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        For the rectangular pulse, it is given by
        $$
            \hat{p}(f) = w \sinc(w f) \mathrm{e}^{-\mathrm{j} 2 \pi (w/2) f}.
        $$

        Examples:
            >>> pulse = komm.RectangularPulse(width=1.0)  # NRZ pulse
            >>> np.abs(pulse.spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )).round(3)
            array([0.   , 0.3  , 0.637, 0.9  , 1.   , 0.9  , 0.637, 0.3  , 0.   ])

            >>> pulse = komm.RectangularPulse(width=0.5)  # Halfway RZ pulse
            >>> np.abs(pulse.spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )).round(3)
            array([0.318, 0.392, 0.45 , 0.487, 0.5  , 0.487, 0.45 , 0.392, 0.318])
        """
        w = self.width
        f = np.asarray(f)
        cexp = np.exp(-2j * np.pi * w / 2 * f)
        centered = w * np.sinc(w * f)
        return centered.astype(complex) * cexp

    def autocorrelation(self, tau: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the rectangular pulse, it is given by
        $$
            R(\tau) = w \tri\left(\frac{\tau}{w}\right).
        $$

        Examples:
            >>> pulse = komm.RectangularPulse(width=1.0)  # NRZ pulse
            >>> pulse.autocorrelation(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )
            array([0.  , 0.25, 0.5 , 0.75, 1.  , 0.75, 0.5 , 0.25, 0.  ])

            >>> pulse = komm.RectangularPulse(width=0.5)  # Halfway RZ pulse
            >>> pulse.autocorrelation(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )
            array([0.  , 0.  , 0.  , 0.25, 0.5 , 0.25, 0.  , 0.  , 0.  ])
        """
        w = self.width
        tau = np.asarray(tau)
        return w * tri(tau / w)

    def energy_density_spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the rectangular pulse, it is given by
        $$
            S(f) = w^2 \sinc^2(w f).
        $$

        Examples:
            >>> pulse = komm.RectangularPulse(width=1.0)  # NRZ pulse
            >>> pulse.energy_density_spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.   , 0.09 , 0.405, 0.811, 1.   , 0.811, 0.405, 0.09 , 0.   ])

            >>> pulse = komm.RectangularPulse(width=0.5)  # Halfway RZ pulse
            >>> pulse.energy_density_spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.101, 0.154, 0.203, 0.237, 0.25 , 0.237, 0.203, 0.154, 0.101])
        """
        w = self.width
        f = np.asarray(f)
        return w**2 * np.sinc(w * f) ** 2

    @cached_property
    def support(self) -> tuple[float, float]:
        r"""
        For the rectangular pulse, the support is given by $[0, w]$.

        Examples:
            >>> pulse = komm.RectangularPulse(width=1.0)
            >>> pulse.support
            (0.0, 1.0)

            >>> pulse = komm.RectangularPulse(width=0.5)
            >>> pulse.support
            (0.0, 0.5)
        """
        return (0.0, self.width)

    def taps(
        self,
        samples_per_symbol: int,
        span: tuple[int, int] | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> pulse = komm.RectangularPulse(width=1.0)
            >>> pulse.taps(samples_per_symbol=4)  # Default span is [0, 1]
            array([1., 1., 1., 1., 0.])
            >>> pulse.taps(samples_per_symbol=4, span=(-1, 1))
            array([0., 0., 0., 0., 1., 1., 1., 1., 0.])
        """
        return super().taps(samples_per_symbol, span)



from abc import ABC, abstractmethod
from functools import cache, cached_property

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from .._util.bit_operations import bits_to_int
from .._util.decorators import blockwise
from ..types import Array1D, Array2D


class BlockCode(ABC):
    @cached_property
    @abstractmethod
    def length(self) -> int:
        r"""
        The length $n$ of the code.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def dimension(self) -> int:
        r"""
        The dimension $k$ of the code.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def redundancy(self) -> int:
        r"""
        The redundancy $m$ of the code.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def rate(self) -> float:
        r"""
        The rate $R = k/n$ of the code.
        """
        return self.dimension / self.length

    @cached_property
    @abstractmethod
    def generator_matrix(self) -> Array2D[np.integer]:
        r"""
        The generator matrix $G \in \mathbb{B}^{k \times n}$ of the code.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def generator_matrix_right_inverse(self) -> Array2D[np.integer]:
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def check_matrix(self) -> Array2D[np.integer]:
        r"""
        The check matrix $H \in \mathbb{B}^{m \times n}$ of the code.
        """
        raise NotImplementedError

    @abstractmethod
    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Applies the encoding mapping $\Enc : \mathbb{B}^k \to \mathbb{B}^n$ of the code. This method takes one or more sequences of messages and returns their corresponding codeword sequences.

        Parameters:
            input: The input sequence(s). Can be either a single sequence whose length is a multiple of $k$, or a multidimensional array where the last dimension is a multiple of $k$.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension expanded from $bk$ to $bn$, where $b$ is a positive integer.
        """

        @blockwise(self.dimension)
        def encode(u: npt.NDArray[np.integer]):
            v = u @ self.generator_matrix % 2
            return v

        return encode(input)

    @abstractmethod
    def project_word(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        @blockwise(self.length)
        def project(v: npt.NDArray[np.integer]):
            u = v @ self.generator_matrix_right_inverse % 2
            return u

        return project(input)

    @abstractmethod
    def inverse_encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Applies the inverse encoding partial mapping $\Enc^{-1} : \mathbb{B}^n \rightharpoonup \mathbb{B}^k$ of the code. This method takes one or more sequences of codewords and returns their corresponding message sequences.

        Parameters:
            input: The input sequence(s). Can be either a single sequence whose length is a multiple of $n$, or a multidimensional array where the last dimension is a multiple of $n$.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension contracted from $bn$ to $bk$, where $b$ is a positive integer.

        Raises:
            ValueError: If the input contains any invalid codewords.
        """
        s = self.check(input)
        if not np.all(s == 0):
            raise ValueError("one or more inputs in 'v' are not valid codewords")
        return self.project_word(input)

    @abstractmethod
    def check(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Applies the check mapping $\mathrm{Chk} : \mathbb{B}^n \to \mathbb{B}^m$ of the code. This method takes one or more sequences of received words and returns their corresponding syndrome sequences.

        Parameters:
            input: The input sequence(s). Can be either a single sequence whose length is a multiple of $n$, or a multidimensional array where the last dimension is a multiple of $n$.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension contracted from $bn$ to $bm$, where $b$ is a positive integer.
        """

        @blockwise(self.length)
        def check(r: npt.NDArray[np.integer]):
            s = r @ self.check_matrix.T % 2
            return s

        return check(input)

    @cache
    @abstractmethod
    def codewords(self) -> Array2D[np.integer]:
        r"""
        Returns the codewords of the code. This is a $2^k \times n$ matrix whose rows are all the codewords. The codeword in row $i$ corresponds to the message obtained by expressing $i$ in binary with $k$ bits (LSB-first).
        """
        k, n = self.dimension, self.length
        batch_size = 1024
        codewords = np.empty((2**k, n), dtype=int)
        for i in tqdm(
            range(0, 2**k, batch_size),
            desc="Generating codewords",
            delay=2.5,
            unit_scale=batch_size,
        ):
            batch_end = min(i + batch_size, 2**k)
            js = np.arange(i, batch_end, dtype=np.uint64).reshape(-1, 1).view(np.uint8)
            messages_batch = np.unpackbits(js, axis=1, count=k, bitorder="little")
            codewords[i:batch_end] = self.encode(messages_batch)
        self._cached_codewords = True
        return codewords

    @cache
    @abstractmethod
    def codeword_weight_distribution(self) -> Array1D[np.integer]:
        r"""
        Returns the codeword weight distribution of the code. This is an array of shape $(n + 1)$ in which element in position $w$ is equal to the number of codewords of Hamming weight $w$, for $w \in [0 : n]$.
        """
        k, n = self.dimension, self.length
        if hasattr(self, "_cached_codewords"):
            return np.bincount(np.sum(self.codewords(), axis=1), minlength=n + 1)
        batch_size = 1024
        distribution = np.zeros(n + 1, dtype=int)
        for i in tqdm(
            range(0, 2**k, batch_size),
            desc="Computing codeword weight distribution",
            delay=2.5,
            unit_scale=batch_size,
        ):
            batch_end = min(i + batch_size, 2**k)
            js = np.arange(i, batch_end, dtype=np.uint64).reshape(-1, 1).view(np.uint8)
            messages_batch = np.unpackbits(js, axis=1, count=k, bitorder="little")
            codewords_batch = self.encode(messages_batch)
            weights = np.sum(codewords_batch, axis=1)
            distribution += np.bincount(weights, minlength=n + 1)
        return distribution

    @cache
    @abstractmethod
    def minimum_distance(self) -> int:
        r"""
        Returns the minimum distance $d$ of the code. This is equal to the minimum Hamming weight of the non-zero codewords.
        """
        return int(np.flatnonzero(self.codeword_weight_distribution())[1])

    @cache
    @abstractmethod
    def coset_leaders(self) -> Array2D[np.integer]:
        r"""
        Returns the coset leaders of the code. This is a $2^m \times n$ matrix whose rows are all the coset leaders. The coset leader in row $i$ corresponds to the syndrome obtained by expressing $i$ in binary with $m$ bits (LSB-first), and whose Hamming weight is minimal. This may be used as a LUT for syndrome-based decoding.
        """
        m, n = self.redundancy, self.length
        H_cols = np.asarray(bits_to_int(self.check_matrix.T))
        visited = np.zeros(2**m, dtype=bool)
        visited[0] = True
        count = 1
        syndromes = [0]
        leaders = np.zeros((2**m, n), dtype=int)
        pbar = tqdm(
            total=2**m,
            desc="Generating coset leaders",
            delay=2.5,
            initial=1,
        )
        while True:
            next_syndromes: list[int] = []
            for s in syndromes:
                for j, h in enumerate(H_cols):
                    s0 = s ^ h
                    if visited[s0]:
                        continue
                    visited[s0] = True
                    leaders[s0] = leaders[s].copy()
                    leaders[s0, j] ^= 1
                    next_syndromes.append(s0)
                    count += 1
                    pbar.update()
                    if count == 2**m:
                        self._cached_coset_leaders = True
                        pbar.close()
                        return leaders
            syndromes = next_syndromes

    @cache
    @abstractmethod
    def coset_leader_weight_distribution(self) -> Array1D[np.integer]:
        r"""
        Returns the coset leader weight distribution of the code. This is an array of shape $(n + 1)$ in which element in position $w$ is equal to the number of coset leaders of weight $w$, for $w \in [0 : n]$.
        """
        m, n = self.redundancy, self.length
        if hasattr(self, "_cached_coset_leaders"):
            return np.bincount(np.sum(self.coset_leaders(), axis=1), minlength=n + 1)
        H_cols = np.asarray(bits_to_int(self.check_matrix.T))
        visited = np.zeros(2**m, dtype=bool)
        visited[0] = True
        count = 1
        syndromes = [0]
        weight = 0
        distribution = np.zeros(n + 1, dtype=int)
        pbar = tqdm(
            total=2**m,
            desc="Computing coset leader weight distribution",
            delay=2.5,
            initial=1,
        )
        while True:
            distribution[weight] = len(syndromes)
            next_syndromes: list[int] = []
            for s in syndromes:
                for h in H_cols:
                    s0 = s ^ h
                    if visited[s0]:
                        continue
                    visited[s0] = True
                    next_syndromes.append(s0)
                    count += 1
                    pbar.update()
                    if count == 2**m:
                        distribution[weight + 1] = len(next_syndromes)
                        pbar.close()
                        return distribution
            syndromes = next_syndromes
            weight += 1

    @cache
    @abstractmethod
    def packing_radius(self) -> int:
        r"""
        Returns the packing radius of the code. This is also called the *error-correcting capability* of the code, and is equal to $\lfloor (d - 1) / 2 \rfloor$.
        """
        return (self.minimum_distance() - 1) // 2

    @cache
    @abstractmethod
    def covering_radius(self) -> int:
        r"""
        Returns the covering radius of the code. This is equal to the maximum Hamming weight of the coset leaders.
        """
        return int(np.flatnonzero(self.coset_leader_weight_distribution())[-1])
